import wx
import numpy as np
from collections import defaultdict

import petram
from petram.utils import get_pkg_datafile
#import petram.geom

fdotbk = get_pkg_datafile(petram.pi, 'icon',  'dot_bk.png')
fedgebk = get_pkg_datafile(petram.pi, 'icon', 'line_bk.png')
ffacebk = get_pkg_datafile(petram.pi, 'icon', 'face_bk.png')
fdot = get_pkg_datafile(petram.pi, 'icon',  'dot.png')
fedge = get_pkg_datafile(petram.pi, 'icon', 'line.png')
fface = get_pkg_datafile(petram.pi, 'icon', 'face.png')
fdom = get_pkg_datafile(petram.pi, 'icon', 'domain.png')
fshow = get_pkg_datafile(petram.pi, 'icon', 'show.png')
fshowall = get_pkg_datafile(petram.pi, 'icon', 'showall.png')
fhide = get_pkg_datafile(petram.pi, 'icon', 'hide.png')
fsolid = get_pkg_datafile(petram.pi, 'icon', 'solid.png')
ftrans = get_pkg_datafile(petram.pi, 'icon', 'transparent.png')


def _select_x(evt, mode, mask):
    viewer = evt.GetEventObject().GetTopLevelParent()
    viewer._sel_mode = mode
    viewer.canvas.unselect_all()
    viewer.set_picker_mask(mask)
    for name, child in viewer.get_axes().get_children():
        if hasattr(child, 'setSelectedIndesx'):
            child.setSelectedIndex([])
    viewer.canvas.unselect_all()
    viewer.draw()


def select_dot(evt):
    _select_x(evt, 'point', 'point')


def select_edge(evt):
    _select_x(evt, 'edge', 'edge')


def select_face(evt):
    _select_x(evt, 'face', 'face')


def select_volume(evt):
    _select_x(evt, 'volume', 'face')


def refresh(navibar, btnls):
    viewer = navibar.GetTopLevelParent()
    mode = viewer._sel_mode
    for btnl in btnls:
        if isinstance(btnl, str):
            continue
        if btnl.tg == 2:
            if btnl.btask == mode:
                btnl.SetToggled(True)
                btnl.SetBitmap(btnl.bitmap2)
            else:
                btnl.SetToggled(False)
                btnl.SetBitmap(btnl.bitmap1)


def show_all(evt):
    viewer = evt.GetEventObject().GetTopLevelParent()
    mode = viewer._sel_mode
    ax = viewer.get_axes()

    namestart = mode if mode != 'volume' else 'face'
    for name, child in ax.get_children():
        if name.startswith('point'):
            continue
        if name.startswith('face'):
            child.hide_component([])
        elif name.startswith('edge'):
            child.hide_component([])

    viewer._hidden_volume = []
    viewer._hidden_face = []
    viewer._hidden_edge = []
    viewer._dom_bdr_sel = ([], [], [], [])
    viewer.draw_all()


def hide_elem(evt, inverse=False):
    viewer = evt.GetEventObject().GetTopLevelParent()
    mode = viewer._sel_mode

    ax = viewer.get_axes()
    namestart = mode if mode != 'volume' else 'face'
    objs = [child for name, child in ax.get_children()
            if name.startswith(namestart)]
    Eobjs = [child for name, child in ax.get_children()
             if name.startswith('edge')]
    Fobjs = [child for name, child in ax.get_children()
             if name.startswith('face')]
    s2l, v2s = viewer._s_v_loop['phys']

    if mode == 'volume':
        facesa = []
        facesb = []
        selected_volume = viewer._dom_bdr_sel[0]

        target_volumes = selected_volume
        if not inverse:
            target_volumes.extend(viewer._hidden_volume)
        for key in v2s.keys():
            if key in target_volumes:
                facesa.extend(v2s[key])
            else:
                facesb.extend(v2s[key])
        if inverse:
            for o in objs:
                o.hide_component(facesa, inverse=True)
            hidden_volume = [x for x in v2s.keys() if not x in selected_volume]
            viewer._hidden_volume = hidden_volume
        else:
            facesa = np.unique(np.array(facesa))
            facesb = np.unique(np.array(facesb))
            new_hide = list(np.setdiff1d(facesa, facesb, True))
            for o in objs:
                idx = o.hidden_component
                idx = list(set(idx+new_hide))
                o.hide_component(idx)

            viewer._hidden_volume.extend(selected_volume)
            viewer._hidden_volume = list(set(viewer._hidden_volume))

    elif mode == 'face' or mode == 'edge':
        for o in objs:
            idx = o.getSelectedIndex()
            if inverse:
                idx = list(set(idx))
            else:
                idx = list(set(o.hidden_component+idx))
            o.hide_component(idx, inverse=inverse)
    elif mode == 'point':
        pass
    else:
        pass

    if mode == 'volume' or mode == 'face':
        from petram.mesh.mesh_utils import line2surf
        hidden_face = sum([o.hidden_component for o in Fobjs], [])
        l2s = line2surf(s2l)

        dd = defaultdict(int)
        for f in hidden_face:
            for x in s2l[f]:
                dd[x] = dd[x]+1
        dd = dict(dd)
        hide_this_edge = [x for x in dd if dd[x] == len(l2s[x])]

        # hide_this_edge = [l for l, ss in l2s.items()
        #              if np.intersect1d(ss, hidden_face).size==len(ss)]

        for o in Eobjs:
            #idx = o.getSelectedIndex()
            idx = list(set(o.hidden_component+hide_this_edge))
            o.hide_component(idx)

    viewer.canvas.unselect_all()
    viewer._dom_bdr_sel = ([], [], [], [])
    viewer.draw_all()


def show_only(evt):
    hide_elem(evt, inverse=True)


def toggle_dot(evt):
    viewer = evt.GetEventObject().GetTopLevelParent()
    mode = viewer._sel_mode

    ax = viewer.get_axes()
    if not ax.has_child('point'):
        return

    if ax.point.isSuppressed:
        ax.point.onUnSuppress()
    else:
        ax.point.onSuppress()
    wx.CallAfter(ax.point.set_gl_hl_use_array_idx, True)
    viewer.canvas.unselect_all()
    # viewer.draw_all()


def toggle_edge(evt):
    viewer = evt.GetEventObject().GetTopLevelParent()
    mode = viewer._sel_mode

    ax = viewer.get_axes()
    if not ax.has_child('edge'):
        return

    if ax.edge.isSuppressed:
        ax.edge.onUnSuppress()
    else:
        ax.edge.onSuppress()

    wx.CallAfter(ax.edge.set_gl_hl_use_array_idx, True)
    viewer.canvas.unselect_all()
    # viewer.draw_all()


def toggle_face(evt):
    viewer = evt.GetEventObject().GetTopLevelParent()
    mode = viewer._sel_mode

    ax = viewer.get_axes()
    children = [(name, child)
                for name, child in ax.get_children() if name.startswith('face')]
    isSuppressed = any([c.isSuppressed for n, c in children])

    if isSuppressed:
        for n, c in children:
            c.onUnSuppress()
            wx.CallAfter(c.set_gl_hl_use_array_idx, True)
    else:
        for n, c in children:
            c.onSuppress()
            wx.CallAfter(c.set_gl_hl_use_array_idx, True)

    viewer.canvas.unselect_all()
    wx.CallAfter(viewer.canvas.refresh_hl)


def make_solid(evt):
    viewer = evt.GetEventObject().GetTopLevelParent()
    mode = viewer._sel_mode

    ax = viewer.get_axes()
    for name, child in ax.get_children():
        if len(child._artists) > 0:
            child.set_alpha(1.0, child._artists[0])
    viewer.canvas.unselect_all()
    viewer.draw_all()


def make_transp(evt):
    viewer = evt.GetEventObject().GetTopLevelParent()
    mode = viewer._sel_mode

    ax = viewer.get_axes()
    for name, child in ax.get_children():
        if len(child._artists) > 0:
            child.set_alpha(0.75, child._artists[0])
    viewer.canvas.unselect_all()
    viewer.draw_all()


btask = [
    ('dot',    fdot,  2, 'select vertex', select_dot),
    ('edge',   fedge, 2, 'select edge', select_edge),
    ('face',   fface, 2, 'select face', select_face),
    ('domain', fdom,  2, 'select domain', select_volume),
    ('---', None, None, None),
    ('toggledot',    fdotbk,  0, 'toggle vertex', toggle_dot),
    ('toggleedge',   fedgebk, 0, 'toggle edge', toggle_edge),
    ('toggleface',   ffacebk, 0, 'toggle face', toggle_face),
    ('---', None, None, None),
    ('mshow',  fshowall,  0, 'show all', show_all),
    ('mhide',  fhide,  0, 'hide selection', hide_elem),
    ('mshowonly',  fshow,  0, 'show only', show_only),
    ('---', None, None, None),
    ('solid',  fsolid,  0, 'solid', make_solid),
    ('transpaent',  ftrans,  0, 'transparent', make_transp), ]



import os
import numpy as np

from petram.model import Model
from .solver_model import Solver
import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('StdSolver')
rprint = debug.regular_print('StdSolver')

class StdSolver(Solver):
    can_delete = True
    has_2nd_panel = False

    def attribute_set(self, v):
        v['clear_wdir'] = False
        v['init_only'] = False   
        v['assemble_real'] = False
        v['save_parmesh'] = False        
        v['phys_model']   = ''
        v['init_setting']   = ''        
        super(StdSolver, self).attribute_set(v)
        return v
    
    def panel1_param(self):
        return [["Initial value setting",   self.init_setting,  0, {},],
                ["physics model",   self.phys_model,  0, {},],
                ["clear working directory",
                 self.clear_wdir,  3, {"text":""}],
                ["initialize solution only",
                 self.init_only,  3, {"text":""}], 
                ["convert to real matrix (complex prob.)",
                 self.assemble_real,  3, {"text":""}],
                ["save parallel mesh",
                 self.save_parmesh,  3, {"text":""}],]

    def get_panel1_value(self):
        return (self.init_setting,
                self.phys_model,
                self.clear_wdir,
                self.init_only,               
                self.assemble_real,
                self.save_parmesh)    
    
    def import_panel1_value(self, v):
        self.init_setting = str(v[0])        
        self.phys_model = str(v[1])
        self.clear_wdir = v[2]
        self.init_only = v[3]        
        self.assemble_real = v[4]
        self.save_parmesh = v[5]        

    def get_editor_menus(self):
        return []
#        return [("Assemble",  self.OnAssemble, None),
#                ("Update RHS",  self.OnUpdateRHS, None),
#                ("Run Solve Step",  self.OnRunSolve, None),]

    def get_phys(self):
        names = self.phys_model.split(',')
        names = [n.strip() for n in names if n.strip() != '']        
        return [self.root()['Phys'][n] for n in names]
    def get_init_setting(self):
        names = self.init_setting.split(',')
        names = [n.strip() for n in names if n.strip() != '']        
        return [self.root()['InitialValue'][n] for n in names]
    
    '''
    This interactive are mostly for debug purpose
    '''
    def OnAssemble(self, evt):
        '''
        assemble linear system interactively (local matrix)
        '''
        dlg = evt.GetEventObject()       
        viewer = dlg.GetParent()
        engine = viewer.engine

        self.assemble(engine)
        self.generate_linear_system(engine)
        evt.Skip()

    def OnUpdateRHS(self, evt):
        dlg = evt.GetEventObject()       
        viewer = dlg.GetParent()
        engine = viewer.engine
        phys = self.get_phys()[0]

        r_B, i_B, extra, r_x, i_x = engine.assemble_rhs(phys, self.is_complex)
        B = engine.generate_rhs(r_B, i_B, extra, r_x, i_x, self.P, format = self.ls_type)
        self.B = [B]
        evt.Skip()

    def OnRunSolve(self, evt):
        dlg = evt.GetEventObject()       
        viewer = dlg.GetParent()
        engine = viewer.engine

        self.call_solver(engine)
        self.postprocess(engine)

    def get_possible_child(self):
        choice = []
        try:
            from petram.solver.mumps_model import MUMPS
            choice.append(MUMPS)
        except ImportError:
            pass

        try:
            from petram.solver.gmres_model import GMRES
            choice.append(GMRES)
        except ImportError:
            pass

        try:
            from petram.solver.strumpack_model import SpSparse
            choice.append(SpSparse)
        except ImportError:
            pass
        return choice
    
    def init_sol(self, engine):
        phys_targets = self.get_phys()
        
        engine.collect_dependent_vars(phys_targets)        
        for p in phys_targets:
            engine.run_mesh_extension(p)
            
        inits = self.get_init_setting()
        if len(inits) == 0:
            # in this case alloate all fespace and initialize all
            # to zero
            engine.run_alloc_sol(phys_targets)
            engine.run_apply_init(phys_targets, 0)
        else:
            for init in inits:
                init.run(engine)
        engine.run_apply_essential(phys_targets)
        return 


    def assemble(self, engine):
        phys_targets = self.get_phys()
        engine.run_verify_setting(phys_targets, self)
        matvecs, matvecs_c = engine.run_assemble(phys_targets, nterms=1)
        return matvecs, matvecs_c

    def generate_linear_system(self, engine, matvecs, matvecs_c):
        phys_target = self.get_phys()
        solver = self.get_active_solver()

        blocks = engine.generate_linear_system(phys_target,
                                            matvecs, matvecs_c)
        # P: projection,  M:matrix, B: rhs, S: extra_flag
        self.M, B, self.Me = blocks
        self.B = [B]

    def store_rhs(self, engine):
        phys_targets = self.get_phys()
        vecs, vecs_c = engine.run_assemble_rhs(phys_targets)
        blocks = engine.generate_rhs(phys_targets, vecs, vecs_c)
        self.B.append(blocks[1])

    def call_solver(self, engine):
        solver = self.get_active_solver()        
        phys_targets = self.get_phys()        
        phys_real = all([not p.is_complex() for p in phys_targets])
        ls_type = solver.linear_system_type(self.assemble_real,
                                            phys_real)
        '''
        ls_type: coo  (matrix in coo format : DMUMP or ZMUMPS)
                 coo_real  (matrix in coo format converted from complex 
                            matrix : DMUMPS)
                 # below is a plan...
                 blk (matrix made mfem:block operator)
                 blk_real (matrix made mfem:block operator for complex
                             problem)
                          (unknowns are in the order of  R_fes1, R_fes2,... I_fes1, Ifes2...)
                 blk_interleave (unknowns are in the order of  R_fes1, I_fes1, R_fes2, I_fes2,...)
                 None(not supported)
        '''
        if debug.debug_memory:
            dprint1("Block Matrix before shring :\n",  self.M)
            dprint1(debug.format_memory_usage())                
        M_block, B_blocks, P = engine.eliminate_and_shrink(self.M,
                                                           self.B, self.Me)
        
        if debug.debug_memory:
            dprint1("Block Matrix after shrink :\n",  M_block)
            dprint1(debug.format_memory_usage())        
        M, B = engine.finalize_linearsystem(M_block, B_blocks,
                                            not phys_real,
                                            format = ls_type)
        solall = solver.solve(engine, M, B)
        #solall = np.zeros((M.shape[0], len(B_blocks))) # this will make fake data to skip solve step
        
        #if ls_type.endswith('_real'):
        if not phys_real and self.assemble_real:
            solall = solver.real_to_complex(solall, M)
        PT = P.transpose()

        return solall, PT

    def store_sol(self, engine, matvecs, solall, PT, ksol = 0):
        phys_targets = self.get_phys()

        sol = PT.reformat_central_mat(solall, ksol)
        sol = PT.dot(sol)

        dprint1(sol)
        l = len(self.B)

        sol, sol_extra = engine.split_sol_array(phys_targets, sol)


        # sol_extra = engine.gather_extra(sol_extra)                

        phys_target = self.get_phys()
        engine.recover_sol(phys_target, matvecs, sol)
        extra_data = engine.process_extra(phys_target, sol_extra)

        return extra_data
            
    def free_matrix(self):
        self.P = None
        self.M = None
        self.B = None

    def save_solution(self, engine, extra_data, 
                      skip_mesh = False, 
                      mesh_only = False):
        phys_target = self.get_phys()
        engine.save_sol_to_file(phys_target, 
                                skip_mesh = skip_mesh,
                                mesh_only = mesh_only,
                                save_parmesh = self.save_parmesh)
        if mesh_only: return
        engine.save_extra_to_file(extra_data)
        engine.is_initialzied = False
        
    def run(self, engine):
        phys_target = self.get_phys()
        if self.clear_wdir:
            engine.remove_solfiles()
        if not engine.isInitialized: self.init_sol(engine)
        if self.init_only:
            extra_data = None
        else:
            matvecs, matvecs_c = self.assemble(engine)
            self.generate_linear_system(engine, matvecs, matvecs_c)
            solall, PT = self.call_solver(engine)
            extra_data = self.store_sol(engine, matvecs, solall, PT, 0)
            dprint1("Extra Data", extra_data)
            
        engine.remove_solfiles()
        dprint1("writing sol files")
        self.save_solution(engine, extra_data)

        print(debug.format_memory_usage())
           






import weakref
from weakref import WeakKeyDictionary as WKD
from weakref import WeakValueDictionary as WVD

class EvaluatorAgent(object):
    def __init__(self):
        object.__init__(self)
        self.mesh = None
        self.knowns = WKD()
        self.emesh_idx = -1
        
    def forget_knowns(self):
        self.knowns = WKD()
        
    def set_mesh(self, mesh):
        self.mesh = weakref.ref(mesh)

    def preprocess_geometry(self, *args, **kwargs):
        raise NotImplementedError("subclass needs to implelment this")

    def eval(self, expr, solvars, phys):    
        raise NotImplementedError("subclass needs to implelment this")        
    



import weakref

import petram.helper.operators as ops
operators = {"integral": ops.Integral,
             "loopintegral": ops.LoopIntegral,
             "identity": ops.Identity,
             "projection": ops.Projection,
             "delta": ops.Delta,
             "deltam": ops.DeltaM,
             "zero": ops.Zero,
             "curl": ops.Curl,
             "grad": ops.Gradient,
             "div": ops.Divergence,
             "convolve": ops.Convolve,
             "hcurln": ops.Hcurln}


def get_operators():
    return {key: operators[key]() for key in operators}

# this is used in vtable to evaluate the expresion containing the operators


def dummy(*args, **kwargs):
    return 1


def get_dummy_operators():
    return {key: dummy for key in operators}


'''
   exp = Expression("integral("domain", "all")")
   exp.set_range_space(...)
   exp.set_test_space(...)
   exp.set_engine(...)
   # exp.set_sel_mode(...)
   # exp.set_sel(...)
   operator = exp.assemble(g = None)

'''


class Expression(object):
    def __init__(self, expr, **kwargs):
        self._expr = expr

        engine = kwargs.pop('engine', None)
        fes1 = kwargs.pop('trial', None)
        fes2 = kwargs.pop('test', None)

        self._sel_mode = kwargs.pop("_sel_mode", "domain")
        self._sel = kwargs.pop("_sel", "all")
        self._fes1 = None
        self._fes2 = None
        self._engine = None
        self._transpose = kwargs.pop('transpose', False)
        # False (horizontal vector if it is vector,,,)
        self._trial_ess_tdof = kwargs.pop('trial_ess_tdof', None)
        self._test_ess_tdof = kwargs.pop('test_ess_tdof', None)
        self._ind_vars = kwargs.pop('ind_vars', 'x, y, z')
        self._is_complex = kwargs.pop('is_complex', False)
        self._c_coeff = kwargs.pop('c_coeff', False)

        super(Expression, self).__init__()
        variables = []
        code = compile(expr, '<string>', 'eval')
        names = code.co_names
        self.co = code

        if fes1 is not None:
            self.set_trial_space(fes1)
        if fes2 is not None:
            self.set_test_space(fes2)
        if engine is not None:
            self.set_engine(engine)

    @property
    def names(self):
        return self._names

    def isOperatorType(self, g):

        from petram.helper.operators import Operator, operators

        # make sure that basic operators are there (if not overwritten)x
        for n in names:
            if n in operators:
                return True

        return False

    def isVariableType(self, g):
        return not isVariableType(g)

    def set_trial_space(self, fes):
        self._fes1 = weakref.ref(fes)

    def set_test_space(self, fes):
        self._fes2 = weakref.ref(fes)

    def set_sel_mode(self, mode):
        self._sel_mode = mode

    def set_sel(self, mode):
        self._sel = mode

    def set_engine(self, en):
        self._engine = weakref.ref(en)

    def assemble(self, g=None):
        g = {} if g is None else g
        operators = get_operators()
        attrs = [
            '_fes1',
            '_fes2',
            '_sel',
            '_sel_mode',
            '_engine',
            '_transpose',
            '_trial_ess_tdof',
            '_test_ess_tdof',
            '_ind_vars',
            '_is_complex',
            '_c_coeff', ]

        for op in operators:
            for attr in attrs:
                setattr(operators[op], attr, getattr(self, attr))
        g2 = g.copy()
        for op in operators:
            g2[op] = operators[op]

        op = eval(self.co, g2)

        return op



from mfem.common.mpi_debug import nicePrint
from petram.phys.vtable import VtableElement, Vtable
from petram.mfem_config import use_parallel
import numpy as np

from petram.model import Domain, Bdry, Edge, Point, Pair
from petram.phys.coefficient import SCoeff, VCoeff
from petram.phys.phys_model import Phys, PhysModule

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('WF_Essential')

if use_parallel:
    import mfem.par as mfem

else:
    import mfem.ser as mfem


class WF_Essential(Bdry, Phys):
    has_essential = True
    nlterms = []
    can_timedpendent = True
    has_3rd_panel = True

    def __init__(self, **kwargs):
        super(WF_Essential, self).__init__(**kwargs)

    def attribute_set(self, v):
        Bdry.attribute_set(self, v)
        Phys.attribute_set(self, v)
        v['sel_readonly'] = False
        v['sel_index'] = []
        v['target_kfes_txt'] = '0'
        return v

    @property
    def vt(self):
        root_phys = self.get_root_phys()
        if not isinstance(root_phys, PhysModule):
            data = (('esse_value', VtableElement(None, type="array",
                                                 guilabel="dummy",
                                                 default="0.0",
                                                 tip="dummy")),)
            return Vtable(data)

        dep_vars = self.get_root_phys().dep_vars
        dep_var = dep_vars[int(self.target_kfes_txt)]

        if not hasattr(self, '_dep_var_bk'):
            self._dep_var_bk = ""

        if self._dep_var_bk != dep_var:
            data = (('esse_value', VtableElement("esse_value", type="array",
                                                 guilabel=dep_var,
                                                 default="0.0",
                                                 tip="Essentail BC")),
                    ('esse_vdim', VtableElement("esse_vdim", type="string",
                                                guilabel="vdim (0-base)",
                                                default="all",
                                                #readonly = True,
                                                tip="vdim (not supported)")),
                    ('esse_elim', VtableElement("esse_elim", type="bool",
                                                guilabel="Use eliminaiton",
                                                default=True,
                                                tip="Eliminate DoFs from linear system (know Essential)")),
                    ('target_kfes', VtableElement("target_kfes", type="string",
                                                  guilabel="FES index",
                                                  default=self.target_kfes_txt,
                                                  tip="target FES")),)

            vt = Vtable(data)
            self._vt1 = vt
            self._dep_var_bk = dep_var
            self.update_attribute_set()
        else:
            vt = self._vt1
        return vt

    def get_essential_idx(self, kfes):
        c0, vdim0, no, target_kfes = self.vt.make_value_or_expression(self)
        target_kfes = np.atleast_1d(eval(target_kfes))

        if kfes in target_kfes:
            return self._sel_index
        else:
            return []

    def apply_essential_1(self, method, real, c0, vdim, vvdim, bdr_attr):
        if vdim == 1:
            coeff1 = SCoeff(c0, self.get_root_phys().ind_vars,
                            self._local_ns, self._global_ns,
                            real=real)
        else:
            coeff1 = VCoeff(vdim, c0, self.get_root_phys().ind_vars,
                            self._local_ns, self._global_ns,
                            real=real)

        assert not (vvdim != -1 and vdim >
                    1), "Wrong setting...(vvdim != -1 and vdim > 1)"
        # print vvdim, vdim, method, coeff1
        if vvdim == -1:
            method(coeff1, mfem.intArray(bdr_attr))
        else:
            for cp in vvdim:
                method(coeff1, mfem.intArray(bdr_attr), cp)

    def apply_essential(self, engine, gf, real=False, kfes=0):
        c0, vdim0, no, target_kfes = self.vt.make_value_or_expression(self)
        target_kfes = np.atleast_1d(eval(target_kfes))

        if kfes not in target_kfes:
            return

        name = self.get_root_phys().dep_vars[kfes]

        if real:
            dprint1("Apply Ess.(real)" +
                    str(self._sel_index), name, 'c0, v0', c0, vdim0)
        else:
            dprint1("Apply Ess.(imag)" +
                    str(self._sel_index), name, 'c0, v0', c0, vdim0)

        fes = engine.get_fes(self.get_root_phys(), name=name)

        vdim = fes.GetVDim()
        vvdim = -1

        if vdim0 != 'all':
            lvdim = len(self.esse_vdim.split(","))
            #assert lvdim == 1, "Specify only one vdim"
            vvdim = [int(x) for x in self.esse_vdim.split(",")]
        else:
            vvdim = -1

        fec_name = fes.FEColl().Name()

        mesh = engine.get_emesh(mm=self)
        ibdr = mesh.bdr_attributes.ToList()

        bdr_attr = [0]*mesh.bdr_attributes.Max()
        for idx in self._sel_index:
            bdr_attr[idx-1] = 1

        if fec_name.startswith("ND"):
            assert vdim == 1, "ND element vdim must be one"
            vdim = mesh.Dimension()
            method = gf.ProjectBdrCoefficientTangent
            self.apply_essential_1(method, real, c0, vdim, vvdim, bdr_attr)

        elif fec_name.startswith("RT"):
            assert vdim == 1, "RT element vdim must be one"
            vdim = mesh.Dimension()
            method = gf.ProjectBdrCoefficientNormal
            self.apply_essential_1(method, real, c0, vdim, vvdim, bdr_attr)

        # elif self.get_root_phys().vdim == 1:
        #    ProjectBdrCoefficient does not realy work in parallel
        #    since shadow vertex are not always set...
        #    method = gf.ProjectBdrCoefficient
        #    self.apply_essential_1(method, real, c0, vdim, vvdim, bdr_attr)

        else:  # H1 or L2
            # vector field FE.
            method = gf.ProjectCoefficient
            ess_vdofs = mfem.intArray()
            fes.GetEssentialVDofs(mfem.intArray(bdr_attr), ess_vdofs, 0)
            vdofs = np.where(np.array(ess_vdofs.ToList()) == -1)[0]
            dofs = mfem.intArray([fes.VDofToDof(i) for i in vdofs])
            fes.BuildDofToArrays()

            if self.get_root_phys().vdim == 1:
                coeff1 = SCoeff(c0, self.get_root_phys().ind_vars,
                                self._local_ns, self._global_ns,
                                real=real)
                gf.ProjectCoefficient(coeff1, dofs, 0)

            elif vdim0 == 'all':
                coeff1 = VCoeff(vdim, c0, self.get_root_phys().ind_vars,
                                self._local_ns, self._global_ns,
                                real=real)
                gf.ProjectCoefficient(coeff1, dofs)
            else:
                for k, cp in enumerate(vvdim):
                    coeff1 = SCoeff(c0, self.get_root_phys().ind_vars,
                                    self._local_ns, self._global_ns,
                                    real=real, component=k)
                    gf.ProjectCoefficient(coeff1, dofs, cp)


class WF_EssentialPoint(Point, Phys):
    '''
    point-wise constraint. works only with H1

    '''
    has_essential = True
    nlterms = []
    can_timedpendent = True
    has_3rd_panel = True

    def __init__(self, **kwargs):
        super(WF_EssentialPoint, self).__init__(**kwargs)

    @classmethod
    def fancy_menu_name(cls):
        return "Essential"

    @classmethod
    def fancy_tree_name(cls):
        return "Essential"

    def attribute_set(self, v):
        Point.attribute_set(self, v)
        Phys.attribute_set(self, v)
        v['sel_readonly'] = False
        v['sel_index'] = []
        v['ess_point_array'] = None
        v['target_kfes_txt'] = '0'
        return v

    def get_ess_point_array(self, kfes):
        c0, vdim0, no, target_kfes = self.vt.make_value_or_expression(self)
        target_kfes = np.atleast_1d(eval(target_kfes))

        if kfes not in target_kfes:
            return []
        return self.ess_point_array

    @property
    def vt(self):
        root_phys = self.get_root_phys()
        if not isinstance(root_phys, PhysModule):
            data = (('esse_value', VtableElement(None, type="array",
                                                 guilabel="dummy",
                                                 default="0.0",
                                                 tip="dummy")),)
            return Vtable(data)

        dep_vars = self.get_root_phys().dep_vars
        dep_var = dep_vars[int(self.target_kfes_txt)]
        if not hasattr(self, '_dep_var_bk'):
            self._dep_var_bk = ""

        if self._dep_var_bk != dep_var:
            data = (('esse_value', VtableElement("esse_value", type="array",
                                                 guilabel=dep_var,
                                                 default="0.0",
                                                 tip="Essentail BC")),
                    ('esse_vdim', VtableElement("esse_vdim", type="string",
                                                guilabel="vdim (0-base)",
                                                default="all",
                                                #readonly = True,
                                                tip="vdim (not supported)")),
                    ('esse_elim', VtableElement("esse_elim", type="bool",
                                                guilabel="Use eliminaiton",
                                                default=True,
                                                tip="Eliminate DoFs from linear system (know Essential)")),
                    ('target_kfes', VtableElement("target_kfes", type="string",
                                                  guilabel="FES index",
                                                  default=self.target_kfes_txt,
                                                  tip="target FES")),)

            vt = Vtable(data)
            self._vt1 = vt
            self._dep_var_bk = dep_var
            self.update_attribute_set()
        else:
            vt = self._vt1
        return vt

    def preprocess_params(self, engine):
        # we use base mesh to process geometry
        mesh = engine.emeshes[0]
        iverts = [mesh.extended_connectivity['vert2vert'][x]
                  for x in self._sel_index]
        ptx = [tuple(mesh.GetVertexArray(iv)) for iv in iverts]
        self.ess_point_array = ptx

    def get_essential_idx(self, kfes):
        c0, vdim0, no, target_kfes = self.vt.make_value_or_expression(self)
        target_kfes = np.atleast_1d(eval(target_kfes))

        if kfes in target_kfes:
            return self._sel_index
        else:
            return []

    def apply_essential(self, engine, gf, real=False, kfes=0):
        c0, vdim0, no, target_kfes = self.vt.make_value_or_expression(self)

        target_kfes = np.atleast_1d(eval(target_kfes))

        if kfes not in target_kfes:
            return

        name = self.get_root_phys().dep_vars[kfes]

        if real:
            dprint1("Apply Ess.(real)" +
                    str(self._sel_index), name, 'c0, v0', c0, vdim0)
        else:
            dprint1("Apply Ess.(imag)" +
                    str(self._sel_index), name, 'c0, v0', c0, vdim0)

        fes = engine.get_fes(self.get_root_phys(), name=name)

        vdim = fes.GetVDim()
        vvdim = -1

        if vdim0 != 'all':
            lvdim = len(self.esse_vdim.split(","))
            #assert lvdim == 1, "Specify only one vdim"
            vvdim = [int(x) for x in self.esse_vdim.split(",")]
        else:
            vvdim = -1

        fec_name = fes.FEColl().Name()
        if not fec_name.startswith("H1"):
            assert False, "Pointwise Essential supports only H1 element"

        dprint1("looking for vertex : ", self.ess_point_array)

        dofs = engine.get_point_essential_tdofs(fes, self.ess_point_array)
        dofs = mfem.intArray(dofs)

        if self.get_root_phys().vdim == 1:
            coeff1 = SCoeff(c0, self.get_root_phys().ind_vars,
                            self._local_ns, self._global_ns,
                            real=real)
            gf.ProjectCoefficient(coeff1, dofs, 0)

        elif vdim0 == 'all':
            coeff1 = VCoeff(vdim, c0, self.get_root_phys().ind_vars,
                            self._local_ns, self._global_ns,
                            real=real)
            gf.ProjectCoefficient(coeff1, dofs)
        else:
            for k, cp in enumerate(vvdim):
                coeff1 = SCoeff(c0, self.get_root_phys().ind_vars,
                                self._local_ns, self._global_ns,
                                real=real, component=k)
                gf.ProjectCoefficient(coeff1, dofs, cp)



from __future__ import print_function

import time
import numpy as np
import weakref
import traceback
import six
from weakref import WeakKeyDictionary as WKD
from weakref import WeakValueDictionary as WVD


from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

import multiprocessing as mp
from petram.sol.evaluators import Evaluator, EvaluatorCommon


class EvaluatorSingle(EvaluatorCommon):
    '''
    define a thing which takes expression involving Vriables
    and evaualte it
    '''
    def __init__(self):
        self.mfem_model = None
        self.solfiles = None
        #self.solvars = WKD()
        self.solvars = {}
        self.agents = {}
        self.physpath = ''
        self.init_done = False
        self.failed = False
        super(EvaluatorSingle, self).__init__()        
        
    def set_solfiles(self, solfiles):
        self.solfiles = solfiles
        # make sure solvars is empty and weakref does not go away.
        self._soliles = solfiles
        self.solvars = {}        
        
    def set_phys_path(self, phys_path):
        self.phys_path = phys_path

    def validate_evaluator(self, name, attr, solfiles, isFirst=False, **kwargs):
        redo_geom = False
        if (self.solfiles is None or 
            self.solfiles.is_different_timestamps(solfiles)):
            self.set_solfiles(solfiles)
            print("new solfiles")
            redo_geom = True
        else:
            print("same solfiles")            
            
        if not super(EvaluatorCommon, self).validate_evaluator(name, attr, **kwargs):
            print("different_setting")
            redo_geom = True
        if not self.init_done: redo_geom = True
        if not redo_geom: return
        
        if isFirst:
            self.init_done = True            
            return

        solvars = self.load_solfiles()
        self.make_agents(self._agent_params[0],
                         attr, **kwargs)

        #for key in list(self.agents):
        #    evaluators = self.agents[key]
        #    for o in evaluators:
        #        o.preprocess_geometry([key], **kwargs)
        self.init_done = True
                
    def eval(self, expr, merge_flag1, merge_flag2, **kwargs):
        if self.phys_path == '': return None, None
        
        phys = self.mfem_model()[self.phys_path]
        solvars = self.load_solfiles()
        if solvars is None: return None, None

        export_type = kwargs.get('export_type', 1)
        
        data = []
        attrs = []
        offset = 0
        def omit_none(l):
            return [x for x in l if x is not None]
        for key in six.iterkeys(self.agents): # scan over battr
            if merge_flag2: offset = 0
            vdata = [] # vertex
            cdata = [] # data
            adata = [] # array idx     
            data.append([])
            attrs.append(key)                                  
            evaluators = self.agents[key]
            for o, solvar in zip(evaluators, solvars): # scan over sol files
                v, c, a = o.eval(expr, solvar, phys, **kwargs)
                if v is None:
                    v = None; c = None; a = None
                else:
                    if merge_flag1: a = a + offset
                    offset = offset + c.shape[0]
                    vdata.append(v)
                    cdata.append(c)
                    adata.append(a)                
                data[-1].append((v, c, a))
            if merge_flag1:
                if len(vdata) != 0:
                    data[-1]  = [(np.vstack(omit_none(vdata)),
                                  np.hstack(omit_none(cdata)),
                                  np.vstack(omit_none(adata)))]
                else:
                    data = data[:-1]  # remove empty tupple
                    attrs = attrs[:-1]
        if export_type == 2: return data, attrs                     
        ## normally this is good option.
        if merge_flag1 and not merge_flag2:
            vdata = np.vstack([x[0][0] for x in data])
            cdata = np.hstack([x[0][1] for x in data])
            adata = np.vstack([x[0][2] for x in data])                                
            data = [(vdata, cdata, adata)]
        elif merge_flag1:
            data0 = []
            for x in data: data0.extend(x)
            data = data0
        elif not merge_flag2:
            keys = self.agents.keys()
            data0 = []
            attr = []
            for idx, o in enumerate(evaluators): # for each file
                vdata = []
                cdata = []
                adata = []
                offset = 0
                for idx0, key in enumerate(keys):
                    d1 = data[idx0][idx]
                    if d1[0] is None: continue
                    vdata.append(d1[0])
                    cdata.append(d1[1])
                    adata.append(d1[2]+offset)
                    offset = offset + d1[1].shape[0]
                if offset == 0: continue
                dd  = (np.vstack(vdata), np.hstack(cdata), np.vstack(adata))
                data0.append(dd)
                attr.append(key)
            attrs = list(set(attr))
            data = data0
        else:
            data0 = []
            for x in data:
                data0.extend([xx for xx in x if xx[0] is not None])
            data = data0
        return data, attrs

    def eval_pointcloud(self, expr, **kwargs):
        if self.phys_path == '': return None, None
        
        phys = self.mfem_model()[self.phys_path]
        solvars = self.load_solfiles()
        if solvars is None: return None, None

        export_type = kwargs.get('export_type', 1)
        
        data = []
        attrs = []
        offset = 0

        key = list(self.agents)[0]
        
        vdata = [] # vertex
        cdata = [] # data
        adata = [] # array idx     
        attrs.append(key)                                  
        evaluators = self.agents[key]
        
        for o, solvar in zip(evaluators, solvars): # scan over sol files
           v, c, a = o.eval(expr, solvar, phys, **kwargs)
           if v is None:
               v = None; c = None; a = None
           else:
               vdata.append(v)
               cdata.append(c)
               adata.append(a)

        if len(adata)==0: return None, None

        ptx  = vdata[0]
        ddim = cdata[0].shape[1:]  # data dim

        shape = adata[0].shape
        shape_d = sum((shape, ddim), ())
                      
        attrs = np.zeros(shape, dtype=int)-1

        #print(shape_d)
        data  = np.zeros(shape_d, dtype=cdata[0].dtype)

        #print("data shape", data.shape)
                
        for v, c, a in zip(vdata, cdata, adata):
            idx = (a != -1)
            if np.sum(idx) == 0: continue
            attrs[idx] = a[idx]
            data[idx] = c
        return ptx, data, attrs

    def eval_integral(self, expr, **kwargs):
        if self.phys_path == '':
            return None, None

        phys = self.mfem_model()[self.phys_path]
        solvars = self.load_solfiles()
        if solvars is None:
            return None

        key = list(self.agents)[0]
        evaluators = self.agents[key]
        
        data = []
        
        v = 0
        num = 0
        for o, solvar in zip(evaluators, solvars): # scan over sol files
            kwargs['num'] = num           
            vv = o.eval_integral(expr, solvar, phys, **kwargs)
            num = num + 1
            if vv is None:
                pass
            else:
                v = v + vv

        return v

    def eval_probe(self, expr, xexpr, probes):
        if self.phys_path == '':
            return None, None, None

        phys = self.mfem_model()[self.phys_path]

        evaluator = self.agents[1][0]
        xval, val = evaluator.eval_probe(expr, xexpr, probes, phys)
        return (0, xval, val)

    def terminate_allnow(self):
        pass



from __future__ import print_function
import wx
import traceback
import numpy as np

import ifigure.widgets.dialog as dialog
from ifigure.utils.edit_list import EditListPanel
from ifigure.utils.edit_list import EDITLIST_CHANGED
from ifigure.utils.edit_list import EDITLIST_CHANGING
from ifigure.utils.edit_list import EDITLIST_SETFOCUS

from petram.pi.simple_frame_plus import SimpleFramePlus

class DlgPlotExpr(SimpleFramePlus):
    def __init__(self, parent, id = wx.ID_ANY, title = 'Plot Expression', **kwargs):
        
        iattr = kwargs.pop("iattr", [])
        
        style = (wx.CAPTION|
                 wx.CLOSE_BOX|
                 wx.MINIMIZE_BOX| 
                 wx.RESIZE_BORDER|
                 wx.FRAME_FLOAT_ON_PARENT|
                 wx.FRAME_TOOL_WINDOW)
        super(DlgPlotExpr, self).__init__(parent, id, title, style=style, **kwargs)
        
        self.parent = parent
        choices = list(parent.model.param.getvar('mfem_model')['Phys'])
        if len(choices) == 0:
            choices = ['No Physics is defined!']

        ll = [['Expression', '', 0, {}],
              ['Index', ','.join(iattr), 0, {}],
              [None, 'Index: Domain (1D/2D) and Boundary (3D)', 2, {}],              
              ['NameSpace', choices[0], 4, {'style':wx.CB_READONLY,
                                            'choices': choices}],
              [None, ' '*50, 2, {}],]
        self.elp = EditListPanel(self, ll)
        button=wx.Button(self, wx.ID_ANY, "Apply")

        box = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(box)        
        box.Add(self.elp, 1, wx.EXPAND|wx.ALL, 1)

        hbox = wx.BoxSizer(wx.HORIZONTAL)        
        hbox.AddStretchSpacer()        
        hbox.Add(button, 0, wx.ALL, 5)
        box.Add(hbox, 0, wx.EXPAND|wx.ALL, 1)

        button.Bind(wx.EVT_BUTTON, self.onApply)
        self.Show()
        self.Layout()
        self.Fit()
        

    def onApply(self, evt):
        value = self.elp.GetValue()
        self._doPlotExpr(value)
        evt.Skip()

    def _doPlotExpr(self, value):
        model = self.parent.model.param.getvar('mfem_model')
        if model is None: return
        
        expr = str(value[0])
        phys = str(value[3])
        use_2dplot = model['Phys'][phys].dim == 1

        from petram.utils import eval_expr
        try:
           engine = self.parent.engine        
           d = eval_expr(model, engine, expr, value[1], phys = phys)
        except:
           dialog.showtraceback(parent = self,
                                txt='Failed to evauate expression',
                                title='Error',
                                traceback=traceback.format_exc())
           return

        
        from ifigure.interactive import figure
        v = figure()
        v.update(False)
        v.suptitle(expr)
        if use_2dplot:
            for k in d.keys():
                v.plot(d[k][0][:, 0, 0].flatten(),
                       d[k][1][:, 0].flatten())                
            v.update(True)            
        else:
            from petram.pi.dlg_plot_sol import setup_figure
            setup_figure(v, self.GetParent())                    
            for k in d.keys():
                v.solid(d[k][0], cz=True, cdata= d[k][1])
            v.update(True)
            v.view('noclip')
            v.lighting(light = 0.5)
        




'''

  variable-table

  a way to define model parameters and to generate associated
  methods

  definition

  name : 'tol'
  guilabel: 'tolelance (%)'
  type : 'any', 'float, 'int', 'complex', 'array', 'string', 'bool'

  special handling note: type == bool
     self.name + '_txt' stores value as boolean
     we use a checkbox interface.

  suffix : [('x', 'y', 'z'), ('x', 'y', 'z')]
  cb :   callback method name
  no_func: True: it can not become variable (use '=')
  tip : tip string
  readonly: text label
  example:

  Scalar : VtableElement('Tbdr',  type='float')
  Vector : VtableElement('Jsurf', type='float', 
                         suffix = ('x', 'y', 'z')
                         default = [0,0,0])
  Matrix : VtableElement('epsilonr', type='complex', 
                         suffix = [('x', 'y', 'z'), ('x', 'y', 'z')]
                         default = np.eye(3,3))

  StaticText :  VtableElement(None,
                              guilabel = 'Default Domain (Vac)',  
                              default =  'eps_r=1, mu_r=1, sigma=0')

  Selectable: define a panel with pull down menu
               (('Z_param', VtableElement('Z_param',
                                   type = 'selectable'
                                   choices= ('Impedance', 'e/m/s'),
                                   vtables = (vt1, vt2)
                                   default = 'Impedance')),)
                       
'''

import six
import numpy as np
import itertools
from collections import OrderedDict

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('Vtable')


def dummy():
    pass


dummy.__repr__ = lambda: '"1"'


class VtableElement(object):
    def __init__(self, name, type='',
                 size=(1,), suffix=None,
                 cb=None, no_func=False,
                 default=0., guilabel=None, tip='',
                 default_txt=None, chkbox=False,
                 choices=None, vtables=None,
                 readonly=False):
        self.name = name
        self.readonly = readonly
        if not isinstance(type, str):
            assert False, "data type should be given as str"
        self.type = type
        self.chkbox = chkbox
        self.choices = choices
        self.vtables = vtables

        if suffix is None:
            self.shape = ()
            self.suffix = []
            self.ndim = 0
        else:
            ndim = 1 if isinstance(suffix[0], str) else 2
            if ndim == 1:
                self.shape = (len(suffix),)
                suffix = (suffix,)
            else:
                self.shape = tuple([len(x) for x in suffix])
            self.ndim = ndim
            self.suffix = [''.join(tmp) for tmp in itertools.product(*suffix)]
        self.cb = cb
        self.no_func = no_func

        if name is not None:
            if type == 'string':
                self.default = default
            elif type == 'array':
                self.default = default
            elif type == 'selectable':
                self.default = default
            elif type == 'bool':
                self.default = bool(default)
            elif type == 'any':
                if default == '':
                    self.default = default
                else:
                    self.default = np.array(default, copy=False)
            else:
                self.default = np.array(default, copy=False)
        else:
            self.default = default
        if len(self.shape) == 0:
            if default_txt is not None:
                self.default_txt = default_txt
            else:
                self.default_txt = self.txt2value(self.default)
        self.guilabel = guilabel if guilabel is not None else self.name
        self.tip = tip

    def txt2value(self, txt):
        # note txt is not always text...
        if self.type == 'float':
            return float(txt)
        elif self.type == 'complex':
            return complex(txt)
        elif self.type == 'int':
            return int(txt)
        elif self.type == 'long':
            return long(txt)
        elif self.type == 'array':
            return txt
        elif self.type == 'string':
            return str(txt)
        elif self.type == 'bool':
            return bool(txt)
        elif self.type == 'any':
            return txt
        elif self.type == '':
            return txt
        elif self.type == 'selectable':
            return txt
        return float(txt)

    def add_attribute(self, v):
        if self.name is None:
            return v

        if self.type == 'bool':
            v[self.name + '_txt'] = self.default

        elif self.type == 'selectable':
            v[self.name + '_select'] = self.default
            for vt in self.vtables:
                v = vt.attribute_set(v)

        elif len(self.shape) == 0:
            v[self.name] = self.default_txt
            v[self.name + '_txt'] = self.default_txt
            if self.chkbox:
                v['use_' + self.name] = False
        else:
            values = [str(x) for x in self.default.flatten()]
            if len(values) != len(self.suffix):
                raise ValueError("Length of defualt value does not match")
            suffix = ['_'+x for x in self.suffix]
            for x, v_txt in zip(suffix, values):
                v[self.name + x] = self.txt2value(v_txt)
                v[self.name + x + '_txt'] = str(v_txt)
            v[self.name + '_m'] = self.default
            xxx = self.default.__repr__().split('(')[1].split(')')[0]
            v[self.name + '_m_txt'] = ''.join(xxx.split("\n"))
            v['use_m_'+self.name] = False
            if self.chkbox:
                v['use_' + self.name] = False
        return v

    def panel_param(self, obj, validator=None):
        if self.name is None:
            return [self.guilabel, self.default,  2, {}]
        chk_float = (self.type == 'float')
        chk_int = (self.type == 'int')
        chk_complex = (self.type == 'complex')
        chk_array = (self.type == 'array')
        chk_string = (self.type == 'string')

        if self.type == 'bool':
            value = getattr(obj, self.name + '_txt')
            ret = [None, value, 3, {"text": self.guilabel}]
            return ret

        elif self.type == 'selectable':
            l = [{'text': self.guilabel, 'choices': self.choices, 'call_fit': False}, ]
            for v in self.vtables:
                elp = v.panel_param(obj)
                l.append({'elp': elp})

            return [None, None, 34, l, ]

        elif len(self.shape) == 0:
            value = getattr(obj, self.name + '_txt')
            ret = obj.make_phys_param_panel(self.guilabel,
                                            value,
                                            no_func=self.no_func,
                                            chk_float=chk_float,
                                            chk_int=chk_int,
                                            chk_complex=chk_complex,
                                            chk_array=chk_array,
                                            chk_string=chk_string)
            if self.readonly:
                ret[2] = ret[2]-10000
            if self.chkbox:
                ret = [None, [True, [value]], 27, [{'text': 'Use'},
                                                   {'elp': [ret]}], ]
            return ret
        else:
            row = self.shape[0]
            col = self.shape[1] if self.ndim == 2 else 1
            suffix = ['_'+x for x in self.suffix]
            ret = obj.make_matrix_panel(self.guilabel, suffix, row=row,
                                        col=col,
                                        validator=validator,
                                        chk_float=chk_float,
                                        chk_int=chk_int,
                                        chk_complex=chk_complex,
                                        no_func=self.no_func)
            if self.readonly:
                ret[2] = ret[2]-10000
            if self.chkbox:
                ret = [None, None, 27, [{'text': 'Use'},
                                        {'elp': [ret]}], ]
            return ret

    def get_panel_value(self, obj):
        if self.name is None:
            return

        if self.type == 'bool':
            value = getattr(obj, self.name + '_txt')
            return value

        elif self.type == 'selectable':
            value = [getattr(obj, self.name + '_select')]
            for vt in self.vtables:
                value.append(vt.get_panel_value(obj))
            return value

        elif len(self.shape) == 0:
            if self.chkbox:
                f = getattr(obj, 'use_' + self.name)
                v = getattr(obj, self.name + '_txt')
                return [f, [v]]
            else:
                return getattr(obj, self.name + '_txt')
        else:
            suffix = ['_'+x for x in self.suffix]
            flag = getattr(obj, 'use_m_'+self.name)
            cb_value = 'Array Form' if flag else 'Elemental Form'
            a = [cb_value,
                 [[str(getattr(obj, self.name+n+'_txt')) for n in suffix]],
                 [str(getattr(obj, self.name + '_m_txt'))]]
            if self.chkbox:
                f = getattr(obj, 'use_' + self.name)
                return [f, [a]]
            else:
                return a

    def import_panel_value(self, obj, v):
        if self.name is None:
            return

        if self.type == 'bool':
            setattr(obj, self.name + '_txt', bool(v))

        elif self.type == 'selectable':
            setattr(obj, self.name + '_select', v[0])
            for k, vt in enumerate(self.vtables):
                vt.import_panel_value(obj, v[k+1])

        elif len(self.shape) == 0:
            if self.chkbox:
                setattr(obj, 'use_' + self.name, v[0])
                setattr(obj, self.name + '_txt', str(v[1][0]))
            else:
                setattr(obj, self.name + '_txt', str(v))
        else:
            if self.chkbox:
                setattr(obj, 'use_' + self.name, v[0])
                v = v[1][0]
            suffix = ['_'+x for x in self.suffix]
            setattr(obj, 'use_m_'+self.name,
                    (str(v[0]) == 'Array Form'))
            for k, n in enumerate(suffix):
                setattr(obj, self.name + n + '_txt', str(v[1][0][k]))
            setattr(obj, self.name + '_m_txt', str(v[2][0]))

    def preprocess_params(self, obj):
        '''
        if no_func, values are evaluated at this stage.
        otherwise, it only makes sure that values are string
        '''
        if self.name is None:
            return

        if self.type == 'bool':
            pass

        elif self.type == 'selectable':
            for v in self.vtables:
                v.preprocess_params(obj)

        elif len(self.shape) == 0:
            if self.no_func:
                kwargs = {}
                kwargs['chk_'+self.type] = True
                v = str(getattr(obj, self.name+'_txt'))
                value = obj.eval_phys_expr(v, self.name, **kwargs)[0]
                setattr(obj, self.name, value)
            else:
                setattr(obj, self.name,
                        str(getattr(obj, self.name + '_txt')))
        else:
            suffix = ['_'+x for x in self.suffix] + ['_m']
            for n in suffix:
                setattr(obj, self.name + n,
                        str(getattr(obj, self.name + n + '_txt')))

    def make_value_or_expression(self, obj):
        if self.name is None:
            return None

        kwargs = {}
        kwargs['chk_'+self.type] = True

        if self.type == 'bool':
            return getattr(obj, self.name+'_txt')

        elif self.type == 'selectable':
            choice = getattr(obj, self.name + '_select')
            idx = self.choices.index(choice)
            return choice, self.vtables[idx].make_value_or_expression(obj)

        elif len(self.shape) == 0:
            if self.no_func:
                return getattr(obj, self.name)
            elif self.type == 'string':
                return str(getattr(obj, self.name))
            else:
                v = getattr(obj, self.name)
                #print('v(before)', v)
                if str(v) == '':
                    v = str(self.default)
                #print('v(after)', v)
                var, f_name0 = obj.eval_phys_expr(v, self.name,
                                                  **kwargs)
                if f_name0 is None:
                    return var
                return f_name0
        else:
            if getattr(obj, 'use_m_'+self.name):
                suffix = ['_m']
                eval_expr = obj.eval_phys_array_expr
                array_mode = True
            else:
                suffix = ['_'+x for x in self.suffix]
                eval_expr = obj.eval_phys_expr
                array_mode = False
            f_name = []
            for n in suffix:
                var, f_name0 = eval_expr(
                    getattr(obj, self.name+n), self.name + n, **kwargs)
                if f_name0 is None:
                    if array_mode:
                        f_name = var
                    else:
                        f_name.append(var)
                else:
                    f_name.append(f_name0)
            #print("f_name here", f_name)
            return f_name

    def panel_tip(self):
        if self.name is None:
            return None
        return self.tip


class Vtable(OrderedDict):
    def attribute_set(self, v, keys=None):
        keys = keys if keys is not None else self.keys()
        for key in keys:
            v = self[key].add_attribute(v)
        return v

    def panel_param(self, obj, keys=None, validator=None):
        keys = keys if keys is not None else self.keys()
        return [self[key].panel_param(obj, validator=validator)
                for key in keys]

    def panel_tip(self, keys=None):
        keys = keys if keys is not None else self.keys()
        return [self[key].panel_tip() for key in keys]

    def get_panel_value(self, obj, keys=None):
        keys = keys if keys is not None else self.keys()
        return [self[key].get_panel_value(obj) for key in keys]

    def import_panel_value(self, obj, values, keys=None):
        keys = keys if keys is not None else self.keys()
        for k, v in zip(keys, values):
            self[k].import_panel_value(obj, v)

    def preprocess_params(self, obj, keys=None):
        keys = keys if keys is not None else self.keys()
        for k in keys:
            self[k].preprocess_params(obj)

    def make_value_or_expression(self, obj, keys=None):
        keys = keys if keys is not None else self.keys()
        return [self[key].make_value_or_expression(obj) for key in keys]


class Vtable_mixin(object):
    def check_phys_expr(self, value, param, ctrl, **kwargs):
        try:
            self.eval_phys_expr(str(value), param, **kwargs)
            return True
        except:
            import petram.debug
            import traceback
            if petram.debug.debug_default_level > 2:
                traceback.print_exc()
            return False

    def check_phys_expr_int(self, value, param, ctrl):
        return self.check_phys_expr(value, param, ctrl, chk_int=True)

    def check_phys_expr_float(self, value, param, ctrl):
        return self.check_phys_expr(value, param, ctrl, chk_float=True)

    def check_phys_expr_complex(self, value, param, ctrl):
        return self.check_phys_expr(value, param, ctrl, chk_complex=True)

    def check_phys_expr_array(self, value, param, ctrl):
        return self.check_phys_expr(value, param, ctrl, chk_array=True)

    def check_phys_array_expr(self, value, param, ctrl, **kwargs):
        try:
            if not 'array' in self._global_ns:
                self._global_ns['array'] = np.array
            self.eval_phys_array_expr(str(value), param, **kwargs)
            return True
        except:
            import petram.debug
            import traceback
            if petram.debug.debug_default_level > 2:
                traceback.print_exc()
            return False

    def check_phys_array_expr_int(self, value, param, ctrl):
        return self.check_phys_array_expr(value, param, ctrl, chk_int=True)

    def check_phys_array_expr_float(self, value, param, ctrl):
        return self.check_phys_array_expr(value, param, ctrl, chk_float=True)

    def check_phys_array_expr_complex(self, value, param, ctrl):
        return self.check_phys_array_expr(value, param, ctrl, chk_complex=True)

    def eval_phys_expr(self, value, param,
                       chk_int=False, chk_complex=False,
                       chk_float=False, chk_array=False,
                       chk_any=False, chk_string=False):

        from petram.helper.variables import NativeCoefficientGenBase
        
        if not hasattr(value, 'startswith'):
            value = str(value)
        
        if value.startswith('='):
            return dummy,  '='.join(value.split('=')[1:])
        else:
            if value.strip() == '':
                return None, None

            g = self._global_ns.copy()
            from petram.helper.expression import get_dummy_operators
            operators = get_dummy_operators()
            for op in operators:
                g[op] = operators[op]

            x = eval(value, g, self._local_ns)

            if isinstance(x, NativeCoefficientGenBase):
                pass
            elif chk_any:
                pass
            elif chk_int:
                x = int(x)
            elif chk_complex:
                x = complex(x)
            elif chk_float:
                x = float(x)
            elif chk_array:
                x = np.atleast_1d(np.array(x, copy=False))
            elif chk_string:
                pass
            else:
                x = x + 0   # at least check if it is number.
            dprint2('Value Evaluation ', param, '=', x)
            return x, None

    def eval_phys_array_expr(self, value, param, chk_complex=False,
                             chk_float=False, chk_int=False):

        from petram.helper.variables import NativeCoefficientGenBase

        if not hasattr(value, 'startswith'):
            value = str(value)
        if value.startswith('='):
            return dummy,  value.split('=')[1]
        else:
            if not 'array' in self._global_ns:
                self._global_ns['array'] = np.array

            if value in self._global_ns:
                x = self._global_ns[value]
                if isinstance(x, NativeCoefficientGenBase):
                    return x, None
            if value in self._local_ns:
                x = self._local_ns[value]
                if isinstance(x, NativeCoefficientGenBase):
                    return x, None
            '''
            make sure that we can make array from input.
            but return an list which is the same as what the elemental 
            form mode would return
            '''
            x = eval('array('+value+')', self._global_ns, self._local_ns)
            if isinstance(x, NativeCoefficientGenBase):
                pass
            else:
                if chk_int:
                    x = x.astype(int)
                elif chk_complex:
                    x = x.astype(complex)
                elif chk_float:
                    x = x.astype(float)
                else:
                    x = x + 0   # at least check if it is number.
                x = list(x.flatten())
            dprint2('Value Evaluation ', param, '=', x)
            return x, None

    # param_panel (defined in NS_mixin) verify if expression can be evaluated
    # phys_param_panel verify if the value is actually float.
    # it forces the number to become float after evaulating the expresison
    # using namespace.
    def make_phys_param_panel(self, base_name, value, no_func=True,
                              chk_int=False,
                              chk_complex=False,
                              chk_float=False,
                              chk_array=False,
                              chk_string=False,
                              validator=None):
        if validator is None:
            if chk_int:
                validator = self.check_phys_expr_int
            elif chk_float:
                validator = self.check_phys_expr_float
            elif chk_complex:
                validator = self.check_phys_expr_complex
            elif chk_string:
                validator = None
            elif chk_array:
                validator = self.check_phys_expr_array
            else:
                validator = self.check_phys_expr

        if no_func:
            return [base_name + "(=)",  value, 0,
                    {'validator': validator,
                     'validator_param': base_name}]
        elif chk_string:
            return [base_name,  value, 0,  {}]
        else:
            return [base_name + "(*)",  value, 0,
                    {'validator':   validator,
                     'validator_param': base_name}]

    def make_matrix_panel(self, base_name, suffix, row=1, col=1,
                          chk_int=False,
                          chk_complex=False,
                          chk_float=False,
                          validator=None,
                          no_func=False):
        if validator is None:

            if chk_int:
                validator = self.check_phys_expr_int
                validatora = self.check_phys_array_expr_int
            elif chk_float:
                validator = self.check_phys_expr_float
                validatora = self.check_phys_array_expr_float
            elif chk_complex:
                validator = self.check_phys_expr_complex
                validatora = self.check_phys_array_expr_complex
            else:
                validator = self.check_phys_expr
                validatora = self.check_phys_array_expr

        a = [{'validator': validator,
              'validator_param': base_name + n} for n in suffix]
        elp1 = [[None, None, 43, {'row': row,
                                  'col': col,
                                  'text_setting': a}], ]
        elp2 = [[None, None, 0, {'validator': validatora,
                                 'validator_param': base_name + '_m'}, ]]

        if no_func:
            label = base_name + '(=) '
        else:
            label = base_name + '(*) '
        ll = [None, None, 34, ({'text': label,
                                'choices': ['Elemental Form', 'Array Form'],
                                'call_fit': False},
                               {'elp': elp1},
                               {'elp': elp2},), ]
        return ll



def make_new_sol():
    if not model.has_child('solutions'): model.add_folder('solutions')

    folder = model.solutions.add_folder('sol')
    folder.mk_owndir()
    param = model.param
    param.setvar('sol', '='+folder.get_full_path())
    return folder

ans(make_new_sol())



'''
  reducer is to eliminate trivial depencencies of 
  blockmatrix

  
  [A  B][x1]  [a1]
  [    ][  ]= [  ] 
  [   C][x2]  [a2]

  
  C x2 = a2 can be solved first.
  
  (1)  When x2 is auxiliary variable, C is often trivial 
           (diagonal, full but small) to invert.
           Then do this first.

  (2)  When x1 is auxiliary variable and A is trivial to invert.
           We can handle this later after x2 is solved by
             A x1 = a1 - B*x2
           Therefore, we dont solve x1 together with x2.

  step1:
       solve (1) and modifiy RHS
  step2:
       remove the row which can be done later by (2)
  step3:
       solve a submatrix which really requres interative solve
  step4:
       solve (2)


  Note: presentely MPI only
'''
from __future__ import print_function

import numpy as np

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('LS_Reducer')

from petram.mfem_config import use_parallel
if use_parallel:
    from petram.helper.mpi_recipes import allgather_vector
    from mfem.common.mpi_debug import nicePrint
    import mfem.par as mfem
   
    from mpi4py import MPI                                   
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank    
else:
    import mfem.ser as mfem
    num_proc = 1
    myid = 0
    def nicePrint(x):
        print(x)

def get_block(Op, i, j):
    try:
        return Op._linked_op[(i,j)]
    except KeyError:
        return None

def do_easy_invert(M):
    '''
    set diagonal element to value (normally 1)
    '''
    col_starts = M.GetColPartArray()
    num_rows, ilower, iupper, jlower, jupper, irn, jcn, data = M.GetCooDataArray()


    from mpi4py import MPI
    myid     = MPI.COMM_WORLD.rank
     
    m = iupper - ilower + 1
    n = jupper - jlower + 1
    n = M.N()

    from scipy.sparse import coo_matrix, lil_matrix
    for i, j in zip(irn, jcn):
        if i != j:
            return -1, None
    try:
        mat =  coo_matrix((1./data, (irn-ilower, jcn)), shape = (m, n)).tolil()
    except:
       print("wrong input")
       print(num_rows, ilower, iupper, jlower, jupper)
       print(np.min(irn-ilower), np.max(irn-ilower) , np.min(jcn),  np.max(jcn), (m, n))
       raise
    from mfem.common.parcsr_extra import ToHypreParCSR
    return  1, ToHypreParCSR(mat.tocsr(), col_starts = col_starts)
    
class LinearSystemReducer(object):
    def __init__(self, opr, name):
        offset = opr.RowOffsets()
        nb = opr.NumRowBlocks()
        self.nb = nb
        
        size = np.diff(offset.ToList())
        lsize = size
        self.lsize = lsize

        if use_parallel:
            size = allgather_vector(size)
            size = np.sum(size.reshape(-1, nb), axis = 0)

        #nicePrint(size)
        filled = np.zeros((nb, nb), dtype=int)
        nnz    = np.zeros((nb, nb), dtype=int)
        for i in range(nb):
           for j in range(nb):
              m = get_block(opr, i, j)
              if m is not None:
                  filled[i, j] = 1
                  nnz[i, j] = m.NNZ()

        flag = [False]*nb
        for i in range(nb): 
            if filled[i, i] == 1 and nnz[i,i] == size[i]:
                flag[i] = True
        #nicePrint("filling pattern", filled)
        #nicePrint("nnz(global)", nnz)
        #nicePrint("easy to invert ", flag)

        easy_inv = [None]*nb
        for i in range(nb):
            if flag[i]:
                check, mat = do_easy_invert(get_block(opr, i, i))
                easy_inv[i] = mat
        self.easy_inv = easy_inv

        first_step_flag = [False]*nb
        for i in range(nb):
            flag = True
            for j in range(nb):
                if  i != j and filled[i, j] == 1:
                    flag = False
            if easy_inv[i] is None: flag = False
            first_step_flag[i] = flag

        last_step_flag = [False]*nb
        for j in range(nb):
            flag = True
            for i in range(nb):
                if filled[i, j] == 1 and i != j:
                    flag = False
            if easy_inv[j] is None: flag = False                
            last_step_flag[j] = flag

        idx = np.logical_or(first_step_flag, last_step_flag)
        reduced_idx = [k for k, n in enumerate(lsize)
                      if not (first_step_flag[k] or last_step_flag[k])]

        self.idx_step1 = list(np.where(first_step_flag)[0])
        self.idx_step2 = list(reduced_idx)
        self.idx_step3 = list(np.where(last_step_flag)[0])
        self.Aname = [n for k, n in enumerate(name) if k in self.idx_step2]
        
        #dprint1("first step: ", self.idx_step1)
        #dprint1("middle step: ", self.idx_step2)        
        #dprint1("last step: ", self.idx_step3)
        #dprint1("A name: ", self.Aname)
        
        rd_offsets = np.hstack([0, np.cumsum(np.array(lsize)[self.idx_step2])])
        self.A = self.make_sub_matrix(opr, rd_offsets, rd_offsets,
                                      self.idx_step2, self.idx_step2)

       
        not_in_step1 = [k for k in range(nb) if not k in self.idx_step1]
        #dprint1("step1", not_in_step1)
        ro = np.hstack([0, np.cumsum(np.array(lsize)[not_in_step1])])
        co = np.hstack([0, np.cumsum(np.array(lsize)[self.idx_step1])])
        self.A_step1 = self.make_sub_matrix(opr, ro, co, 
                                            not_in_step1, self.idx_step1)

        not_in_step3 = [k for k in range(nb) if not k in self.idx_step3]
        #dprint1("step3", not_in_step3)        
        ro = np.hstack([0, np.cumsum(np.array(lsize)[self.idx_step3])])
        co = np.hstack([0, np.cumsum(np.array(lsize)[not_in_step3])])
        #dprint1("step3", ro, co)                
        self.A_step3 = self.make_sub_matrix(opr, ro, co, 
                                            self.idx_step3, not_in_step3)

        self.Ainv_step1 = self.make_inv_mat(self.idx_step1, easy_inv)
        self.Ainv_step3 = self.make_inv_mat(self.idx_step3, easy_inv)
        
    def make_sub_matrix(self, src_mat, roffset, coffset, row_idx, col_idx):
        ro = mfem.intArray(list(roffset))
        co = mfem.intArray(list(coffset))
        mat = mfem.BlockOperator(ro, co)
        for ii, i in enumerate(row_idx):
           for jj, j in enumerate(col_idx):
               m = get_block(src_mat, i, j)
               if m is None: continue
               mat.SetBlock(ii, jj,  m)
        return mat

    def make_inv_mat(self, idx, inv):
        ro = np.hstack([0, np.cumsum(np.array(self.lsize)[idx])])
        row = mfem.intArray(list(ro))
        col = mfem.intArray(list(ro))
        mat = mfem.BlockOperator(row, col)
        
        for ii, i in enumerate(idx):
               mat.SetBlock(ii, ii,  inv[i])
        return mat

    def make_sub_vec(self, src_vec, idx, inv=False, nocopy=False):
        if inv:
            idx = [k for k in range(self.nb) if not k in idx]
        size = [src_vec.BlockSize(i) for i in idx]
        offset = np.hstack([0, np.cumsum(size, dtype=int)])
        offset = mfem.intArray(list(offset))
        
        sub_vec = mfem.BlockVector(offset)
        sub_vec._offsets = offset # in order to keep it from freed

        for k, i in enumerate(idx):
           if nocopy:
               sub_vec.GetBlock(k).Assign(0.0)
           else:
               sub_vec.GetBlock(k).Assign(src_vec.GetBlock(i).GetDataArray())

        return sub_vec

    def set_solver(self, solver):
        self.solver = solver
        
    def Mult(self, bb, xx, assert_convergence = False):
        bb_1 = self.make_sub_vec(bb, self.idx_step1)
        xx_1 = self.make_sub_vec(xx, self.idx_step1)        
        self.Ainv_step1.Mult(bb_1, xx_1)


        for k, i in enumerate(self.idx_step1):
            xx.GetBlock(i).Assign(xx_1.GetBlock(k).GetDataArray())

        #nicePrint("xx_1", xx_1.GetBlock(0).GetDataArray())
        
        bb_1_2 = self.make_sub_vec(bb, self.idx_step1, inv=True, nocopy=True)

        self.A_step1.Mult(xx_1, bb_1_2)

        for k, i in enumerate(self.idx_step2+self.idx_step3):
            v = bb.GetBlock(i)
            v -= bb_1_2.GetBlock(k)

        bb_2 = self.make_sub_vec(bb, self.idx_step2)
        xx_2 = self.make_sub_vec(xx, self.idx_step2)

        self.solver.Mult(bb_2, xx_2)
        max_iter = self.solver.GetNumIterations();
        tol = self.solver.GetFinalNorm()
        
        dprint1("convergence check (max_iter, tol) ", max_iter, " ", tol)
        if assert_convergence:
            if not self.solver.GetConverged():
                raise debug.ConvergenceError("Convergence error when solving Reduced linear system")

        for k, i in enumerate(self.idx_step2):
            xx.GetBlock(i).Assign(xx_2.GetBlock(k).GetDataArray())


        xx_3_1 = self.make_sub_vec(xx, self.idx_step3, inv=True)
        xx_3_2 = self.make_sub_vec(xx, self.idx_step3)
        bb_3 = self.make_sub_vec(bb, self.idx_step3)
        self.A_step3.Mult(xx_3_1, bb_3)

        for k, i in enumerate(self.idx_step3):
            v = bb.GetBlock(i)
            v -= bb_3.GetBlock(k)
        bb_3 = self.make_sub_vec(bb, self.idx_step3)            
        self.Ainv_step3.Mult(bb_3, xx_3_2)
        
        for k, i in enumerate(self.idx_step3):
            xx.GetBlock(i).Assign(xx_3_2.GetBlock(k).GetDataArray())
        
        
        

        
        
                
        
               


            




'''
   extract_refined_mesh_data3

   extract refined  mesh data from 2D (=ndim)  mesh
'''
import numpy as np
from collections import defaultdict

from petram.mesh.find_edges import find_edges
from petram.mesh.find_vertex import find_vertex
from mfem.ser import GlobGeometryRefiner as GR

def extract_refined_mesh_data2(mesh, refine=None):
    ndim = mesh.Dimension()
    sdim = mesh.SpaceDimension()
    
    ivert0 = [mesh.GetElement(i).GetVerticesArray()
                           for i in range(mesh.GetNE())]
    attrs = mesh.GetAttributeArray()
    nvert = np.array([len(x) for x in ivert0])
    idx3 = np.where(nvert == 3)[0]
    idx4 = np.where(nvert == 4)[0]
    ivert = []; ivert3 = None; ivert4 = None
    iv3 = []; iv4 = []
    iv3p = np.array([], dtype=int); iv4p = np.array([], dtype=int)

    from petram.mesh.refined_mfem_geom import get_geom

    ptx = []; ivx3 = None; ptx3 = []
    if len(idx3) != 0:
        base = mesh.GetElementBaseGeometry(idx3[0])
        gt = mesh.GetElementTransformation        
        attr3, ptx3, ivx3, ivxe3, attrx3 =  get_geom(idx3, 3, base,
                                                     gt, attrs, sdim, refine)
        ptx.append(ptx3)
        npt = len(ptx3)//len(idx3)        
        lref = len(ivx3)//len(idx3)
        iv3  = [ivert0[k] for k in idx3]
        seen = defaultdict(int)
        for iiv in ivx3[:lref].flatten():
            seen[iiv] += 1
        xxx = np.sort([kk for kk in seen if seen[kk]==1])
        iv3p = np.hstack([xxx+j*npt for j, k in enumerate(idx3)])
        ivert3 = np.vstack(iv3)
    if len(idx4) != 0:
        base = mesh.GetElementBaseGeometry(idx4[0])
        gt = mesh.GetElementTransformation                
        attr4, ptx4, ivx4, ivxe4, attrx4 =  get_geom(idx4, 4, base,
                                                          gt, attrs,sdim,refine)
        ptx.append(ptx4)
        npt = len(ptx4)//len(idx4)
        lref = len(ivx4)//len(idx4)
        iv4 = [ivert0[k] for k in idx4]        
        seen = defaultdict(int)
        for iiv in ivx4[:lref].flatten():
            seen[iiv] += 1
        xxx = np.sort([kk for kk in seen if seen[kk]==1])
        iv4p = np.hstack([xxx+j*npt for j, k in enumerate(idx4)])+len(ptx3)        
        ivert4 = np.vstack(iv4)
        if ivx3 is not None:
            ivx4 = ivx4 + len(ptx3)       
            ivxe4 = ivxe4 + len(ptx3)
        
    ptx_face = np.vstack(ptx)
    
    # unique edges for node

    tmp = np.hstack(iv3 + iv4)
    tmp2 = np.hstack((iv3p, iv4p))
    u, indices = np.unique(tmp, return_inverse=True)
    
    ll3 = 3*len(idx3)
    indices3 = indices[:ll3]
    indices4 = indices[ll3:]
                       
    table = np.zeros(np.max(u)+1, dtype=int)-1
    for k, u0 in enumerate(u):
        table[u0] = tmp2[np.where(tmp == u0)[0][0]]

    X = ptx_face
    #X = np.vstack([mesh.GetVertexArray(k) for k in u])    
    if X.shape[1] == 2:
        X = np.hstack((X, np.zeros((X.shape[0],1))))
    elif X.shape[1] == 1:
        X = np.hstack((X, np.zeros((X.shape[0],2))))
    
    kdom = mesh.attributes.ToList()
    
    cells = {}
    cell_data = {}
    cell_data['X_refined_face']=ptx_face
    if ivert3 is not None:
        cells['triangle'] = indices3.reshape(ivert3.shape)
        cells['triangle_x'] = ivx3
        cells['triangle_xe'] = ivxe3        
        cell_data['triangle'] = {}
        cell_data['triangle']['physical'] = attr3
        cell_data['triangle_x'] = {}                         
        cell_data['triangle_x']['physical'] = attrx3
    if ivert4 is not None:
        cells['quad'] = indices4.reshape(ivert4.shape)
        cells['quad_x'] = ivx4
        cells['quad_xe'] = ivxe4        
        cell_data['quad'] = {}
        cell_data['quad']['physical'] = attr4
        cell_data['quad_x'] = {}                         
        cell_data['quad_x']['physical'] = attrx4
        
    ### process refined edges
    battrs = mesh.GetBdrAttributeArray()        
    base = 1
    gt = mesh.GetBdrElementTransformation
    idx2 = range(mesh.GetNBE())
    if len(idx2) > 0:
        attr2, ptx2, ivx2, ivxe2, attrx2 = get_geom(idx2, 2, base, gt, battrs,
                                                sdim, refine)
        cells['line_x'] = ivx2
        cell_data['line_x'] = {}                        
        cell_data['line_x']['physical'] = attrx2
        cell_data['X_refined_edge']=ptx2
    
    
    from petram.mesh.mesh_utils import populate_plotdata
    l_s_loop = populate_plotdata(mesh, table, cells, cell_data)
    iedge2bb = None # is it used?    

    #print "X", X.shape
    #for k in cells['vertex']:print X[k]
    return X, cells, cell_data, l_s_loop, iedge2bb





import os
from os.path import expanduser, abspath
import shutil
import wx
import time

from ifigure.utils.mp_tarzip import MPTarzip
'''
  usage

  from petram.pi.pfz2script import generate, generate_all

  path_src = '~/piscope_projects/PetraM_test/piscope_projects'
  path_dst = '~/src/TwoPiTest/PetraM_test'
  generate(path_src, path_dst, 'em3d_TEwg.pfz')
'''

def generate(path_src, path_dst, f, create_new = True, reload_scripts=False):
    #  script generater 
    #    create_new : start a new project and load the project file
    #
    path_src = abspath(expanduser(path_src))
    path_dst = abspath(expanduser(path_dst))
    
    app = wx.GetApp().TopWindow
    if create_new: app.onNew(e=None, use_dialog=False)
    app.onOpen(path=os.path.join(path_src, f))
    proj = app.proj
    model = proj.setting.parameters.eval("PetraM")

    from petram.pi.shell_commands import import_project_scripts
    if reload_scripts:
        scripts = model.scripts
        for name, child in scripts.get_children():
            child.destroy()
        scripts.clean_owndir()
        import_project_scripts(scripts)
        app.onSave(None)
        while not MPTarzip().isReady():
            time.sleep(3)
            
    
    m = proj.model1.mfem.param.eval("mfem_model")
    m.set_root_path(model.owndir())
    
    for od in m.walk():
        if hasattr(od, 'use_relative_path'):
            od.use_relative_path()
            
    dst = os.path.join(path_dst, f.split(".")[0])
    if not os.path.exists(dst):
        os.mkdir(dst)
    m.generate_script(dir=dst)

    for od in m.walk():
        if hasattr(od, 'use_relative_path'):
            od.restore_fullpath()
            mesh_path = od.get_real_path()
            mesh_path2 = os.path.join(dst, os.path.basename(od.path))
            if os.path.exists(mesh_path2): continue
            if os.path.exists(mesh_path): shutil.copy(mesh_path, mesh_path2)

def generate_all(path_src, path_dst):
    '''
     process all pfz file in a directoy

    '''
    path_src = abspath(expanduser(path_src))    
    files = os.listdir(path_src)
    create_new = True
    for f in files:
        if not f.endswith('.pfz'): continue
        print("working on ", f)
        generate(path_src, path_dst, f, create_new=create_new)
        create_new = False



from petram.helper.matrix_file import write_coo_matrix, write_vector
import sys
import os
import numpy as np
import scipy.sparse
from collections import OrderedDict
from warnings import warn

from petram.mfem_config import use_parallel
if use_parallel:
    from petram.helper.mpi_recipes import *
    import mfem.par as mfem
else:
    import mfem.ser as mfem
import mfem.common.chypre as chypre

# these are only for debuging
from mfem.common.parcsr_extra import ToScipyCoo
from mfem.common.mpi_debug import nicePrint

from petram.phys.phys_model import Phys
from petram.model import Domain, Bdry, ModelDict
from petram.phys.vtable import VtableElement, Vtable


import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('AUX_Variable')

#groups = ['Domain', 'Boundary', 'Edge', 'Point', 'Pair']
groups = ['Domain', 'Boundary', 'Pair']

data0 = [("oprt_diag", VtableElement("oprt_diag", type='array',
                                     guilabel="diag", default="0.0",
                                     tip="oprator (diag)",)),
         ("rhs_vec", VtableElement("rhs_vec", type='array',
                                   guilabel="rhs", default="0.0",
                                   tip="rhs vector",))]


class AUX_Variable(Phys):
    vt_diag_rhs = Vtable(data0)
    has_3rd_panel = True
    _has_4th_panel = True
    extra_diagnostic_print = True

    def attribute_set(self, v):
        v = super(AUX_Variable, self).attribute_set(v)
        v["variable_name"] = ""
        v["aux_connection"] = OrderedDict({})
        v["jmatrix_config"] = None
        v = self.vt_diag_rhs.attribute_set(v)

        if not hasattr(self, '_vt_array'):
            self._vt_array = []
        for vt in self.vt_array:
            v = vt.attribute_set(v)
            #vv = vt.attribute_set({})
            # for key in vv:
            #    if hasattr(self, key): vv[key] = getattr(self, key)
            #    v[key] = vv[key]
        return v

    def save_attribute_set(self, skip_def_check):
        attrs = super(AUX_Variable, self).save_attribute_set(skip_def_check)
        for vt in self.vt_array:
            vv = vt.attribute_set({})
            for key in vv:
                if not key in attrs:
                    attrs.append(key)
        return attrs

    def extra_DoF_name(self):
        return self.variable_name

    @property
    def vt_array(self):
        if not hasattr(self, 'aux_connection'):
            self._vt_array = []
            return self._vt_array

        if not hasattr(self, '_vt_array'):
            self._vt_array = []
        for key in self.aux_connection:
            if len(self._vt_array) > key:
                continue
            sidx = str(key)
            data = [("oprt1_"+sidx, VtableElement("oprt1_"+sidx, type='any',
                                                  guilabel="operator1", default="",
                                                  tip="oprator (horizontal)",)),
                    ("oprt2_"+sidx, VtableElement("oprt2_"+sidx, type='any',
                                                  guilabel="operator2", default="",
                                                  tip="oprator (vertical)",)), ]
            vt = Vtable(data)
            self._vt_array.append(vt)
        return self._vt_array

    def panel1_param(self):
        import wx
        from wx import BU_EXACTFIT

        b1 = {"label": "+", "func": self.onAddConnection,
              "noexpand": True, "style": BU_EXACTFIT}  # , "sendevent":True}
        b2 = {"label": "-", "func": self.onRmConnection,
              "noexpand": True, "style": BU_EXACTFIT}  # , "sendevent":True}

        ll = [["name"+" "*16,   self.variable_name, 0, {}], ]

        ll.extend(self.vt_diag_rhs.panel_param(self))
        ll.append([None, None, 241, {'buttons': [b1, b2],
                                     'alignright':True,
                                     'noexpand': True}])

        mfem_physroot = self.get_root_phys().parent
        names, pnames, pindex = mfem_physroot.dependent_values()
        names = [n+" ("+p + ")" for n, p in zip(names, pnames)]

        '''
        if not hasattr(self, '_vt_array'): self._vt_array = []        
        for key in self.aux_connection:
            if len(self._vt_array) > key: continue
            sidx = str(key)
            data = [("oprt1_"+sidx, VtableElement("oprt1_"+sidx, type='any',
                                           guilabel = "operator1", default = "",
                                           tip = "oprator (horizontal)",)),
                    ("oprt2_"+sidx, VtableElement("oprt2_"+sidx, type='any',
                                           guilabel = "operator2", default = "",
                                           tip = "oprator (vertical)",)),]
            vt = Vtable(data)
            self._vt_array.append(vt)
        '''
        self.update_attribute_set()
        for j, key in enumerate(self.aux_connection):
            ll1 = [["paired variable", "S", 4,
                    {"style": wx.CB_READONLY, "choices": names}]]
            ll2 = self._vt_array[j].panel_param(self)
            ll.extend(ll1 + ll2)

        return ll

    def import_panel1_value(self, v):
        mfem_physroot = self.get_root_phys().parent
        names, pnames, pindex = mfem_physroot.dependent_values()

        if len(str(v[0])) == 0:
            dprint1("Name of variable must be given")
        self.variable_name = str(v[0])
        self.vt_diag_rhs.import_panel_value(self, v[1:3])
        i_st = 4
        for i, key in enumerate(self.aux_connection):

            idx = names.index(str(v[i_st]).split("(")[0].strip())
            self.aux_connection[key] = (pnames[idx], pindex[idx])
            # if len(self._vt_array) >= i: continue
            self._vt_array[i].import_panel_value(self,
                                                 v[(i_st+1):(i_st+3)])
            i_st = i_st + 3

    def get_panel1_value(self):
        def get_label(pair):
            if pair is None:
                n = self.get_root_phys().dep_vars[0]
                p = self.get_root_phys().name()
            else:
                mfem_physroot = self.get_root_phys().parent
                var_s = mfem_physroot[pair[0]].dep_vars
                n = var_s[pair[1]]
                p = pair[0]
            var = n + " ("+p + ")"
            return var

        v = [self.variable_name, ]
        v.extend(self.vt_diag_rhs.get_panel_value(self))
        v.append(None)
        for i, key in enumerate(self.aux_connection):
            v.append(get_label(self.aux_connection[key]))
            v.extend(self._vt_array[i].get_panel_value(self))

        return v

    def panel2_param(self):
        return [[None, "Auxiriary varialbe is global",  2,   {}], ]

    def import_panel2_value(self, v):
        pass

    def get_panel2_value(self):
        return [None]

    def panel3_param(self):
        return [[None, "Auxiriary varialbe is linear/no init.",  2,   {}], ]

    def import_panel3_value(self, v):
        pass

    def get_panel3_value(self):
        return [None]

    def panel4_param(self):
        ll = super(AUX_Variable, self).panel4_param()
        ll.append(['Varying (in time/for loop) RHS', False, 3, {"text": ""}])
        return ll

    def panel4_tip(self):
        return None

    def import_panel4_value(self, value):
        super(AUX_Variable, self).import_panel4_value(value[:-1])
        self.isTimeDependent_RHS = value[-1]

    def get_panel4_value(self):
        value = super(AUX_Variable, self).get_panel4_value()
        value.append(self.isTimeDependent_RHS)
        return value

    def check_extra_update(self, mode):
        '''
        mode = 'B' or 'M'
        'M' return True, if M needs to be updated
        'B' return True, if B needs to be updated
        '''
        if self._update_flag:
            if mode == 'B':
                return self.isTimeDependent_RHS
            if mode == 'M':
                return self.isTimeDependent
        return False

    def onAddConnection(self, evt):
        mfem_physroot = self.get_root_phys().parent
        names, pnames, pindex = mfem_physroot.dependent_values()
        names = [n+" ("+p + ")" for n, p in zip(names, pnames)]

        keys = self.aux_connection.keys()
        if len(keys) == 0:
            self.aux_connection[0] = (pnames[0], 0)
        else:
            self.aux_connection[max(keys)+1] = (pnames[0], 0)
        evt.GetEventObject().TopLevelParent.OnItemSelChanged()

    def onRmConnection(self, evt):
        if len(self._vt_array) < 1:
            return
        keys = self.aux_connection.keys()
        del self.aux_connection[keys[-1]]
        self._vt_array = self._vt_array[:-1]
        evt.GetEventObject().TopLevelParent.OnItemSelChanged()

    def has_extra_DoF2(self, kfes, phys, jmatrix):
        if not self.timestep_config[jmatrix]:
            return False

        flag = False
        for key in self.aux_connection:
            phys_name, kkfes = self.aux_connection[key]
            if (phys.name() == phys_name) and (kfes == kkfes):
                flag = True
            if flag:
                return True
        return False

    def postprocess_extra(self, sol, flag, sol_extra):
        name = self.variable_name
        sol_extra[name] = sol.toarray()

    def preprocess_params(self, engine):
        self.vt_diag_rhs.preprocess_params(self)
        for vt in self._vt_array:
            vt.preprocess_params(self)
        super(AUX_Variable, self).preprocess_params(engine)

    def _add_vt_array(self, sidx):
        data = [("oprt1_"+sidx, VtableElement("oprt1_"+sidx, type='any',
                                              guilabel="operator1", default="",
                                              tip="oprator (horizontal)",)),
                ("oprt2_"+sidx, VtableElement("oprt2_"+sidx, type='any',
                                              guilabel="operator2", default="",
                                              tip="oprator (vertical)",)), ]
        vt = Vtable(data)
        self._vt_array.append(vt)
        vt.preprocess_params(self)

    def add_extra_contribution(self, engine, ess_tdof=None,
                               kfes=0, phys=None):
        dprint1("Add Extra contribution : " + self.name())
        range = self.get_root_phys().dep_vars[kfes]

        diag, rhs_vec = self.vt_diag_rhs.make_value_or_expression(self)

        if isinstance(rhs_vec, str):
            g = self._global_ns.copy()
            l = self._local_ns.copy()
            rhs_vec = np.atleast_1d(eval(rhs_vec, g, l)).flatten()
            dprint1("rhs after evaluation", rhs_vec)

        if isinstance(diag, str):
            g = self._global_ns.copy()
            l = self._local_ns.copy()
            diag = np.atleast_1d(eval(diag, g, l)).flatten()
            dprint1("diag after evaluation", diag)

        for i, key in enumerate(self.aux_connection):
            if len(self._vt_array) < i+1:
                self._add_vt_array(str(key))
            opr1, opr2 = self._vt_array[i].make_value_or_expression(self)

        t1 = None
        t2 = None
        t3 = None
        t4 = None

        from petram.helper.expression import Expression

        name = phys.dep_vars[kfes]
        fes = engine.fespaces[name]

        ind_vars = self.get_root_phys().ind_vars
        is_complex = self.get_root_phys().is_complex()

        diag_size = -1
        if opr1 is not None or opr2 is not None:
            if opr1 is not None:
                assert isinstance(opr1, str), "operator1 must be an expression"
                expr = Expression(opr1, engine=engine, trial=fes,
                                  ind_vars=ind_vars, is_complex=is_complex)
                t1 = expr.assemble(g=self._global_ns)
                diag_size = t1.shape[0]
            if opr2 is not None:
                assert isinstance(opr2, str), "operator2 must be an expression"
                expr = Expression(opr2, engine=engine, trial=fes, transpose=True,
                                  ind_vars=ind_vars, is_complex=is_complex)
                t2 = expr.assemble(g=self._global_ns)
                if diag_size > -1:
                    # print t1.shape, t2.shape
                    assert diag_size == t2.shape[1], "t1 and t2 shapes are inconsistent"
                diag_size = t2.shape[1]

        from mfem.common.chypre import IdentityPyMat, Array2PyVec
        if diag is not None:
            if not self.get_root_phys().is_complex():
                diag = diag.real
            t3 = IdentityPyMat(diag_size, diag=diag)

        # all node does do the same job, but Array2PyVec will use the data from
        # root node.

        rhs = np.atleast_1d(rhs_vec).astype(float)
        if diag_size != 1 and len(rhs) == 1:
            rhs = np.hstack([rhs[0]]*diag_size).flatten()

        if self.get_root_phys().is_complex():
            rhs = rhs.astype(complex, copy=False)
        if not self.get_root_phys().is_complex():
            if np.iscomplexobj(rhs):
                rhs = rhs.real

        t4 = Array2PyVec(rhs)

        '''
        Format of extra.
        [M,  t2]   [  ]
        [      ] = [  ]
        [t1, t3]   [t4]

        and it returns if Lagurangian will be saved.

        '''
        return (t2, t1, t3, t4, True)



def save_model(path, meshfile_relativepath=False, do_preprocess = False):
    od = model.param.eval('mfem_model')
    if do_preprocess:
        model.scripts.helpers.rebuild_ns()        
        engine.run_preprocess(model.namespaces, model.datasets)
    od.save_to_file(path, meshfile_relativepath=meshfile_relativepath)
    model.variables.setvar('modelfile_path', path)

ans(save_model(*args, **kwargs))



import os
import shutil
import petram.helper.pickle_wrapper as pickle

from os.path import expanduser

def save_test(model):
    from ifigure.widgets.dialog import writedir
    path = writedir(message='Enter Save dirctory')
    if path == '': return
    path = expanduser(path)

    for od in model.walk():
         if hasattr(od, 'use_relative_path'):
             mpath = od.get_real_path()             
             od.use_relative_path()
             shutil.copyfile(mpath,
                             os.path.join(path, os.path.basename(mpath)))

    try:
        pickle.dump(model, open(os.path.join(path, 'model.pmfm'),
                                'wb'))        
        model.generate_script(dir = path, 
                          parallel = False,
                          filename = 'model.py')

        model.generate_script(dir = path, 
                          parallel = True,
                          filename = 'modelp.py')
    except:
        import traceback
        traceback.print_exc()

    for od in model.walk():
       if hasattr(od, 'use_relative_path'):
            od.restore_fullpath()
    



from __future__ import print_function
import resource
import time
import textwrap
#####################################
#
# debug.py
#
#    provide a simple interface to make debug prints
#    allow for controling turn on and off of debug
#    print for each module
#
# Usage:
#    (in import section)
#    import ifigure.utils.debug as debug
#    dprint1, dprint2, dprint3 = debug.init_dprints('ArgsParser', level=0)
#
#    (then use it as follows)
#    debug.set_level('ArgsParser', 1)  # set level for ArsgParser 1
#    dprint1('hogehogehoge')           # print something
#
#    level 1 (dprint1) : usr feedback which will be turn on normally
#    level 2 (dprint2) : first level of debug print
#    level 3 (dprint3) : second level of debug print
#    setting debug_default_level to 0 will turn off all error print
#    (silent mode)


import traceback
import six

debug_mode = 1
debug_modes = {}
debug_default_level = 1
debug_essential_bc = False
debug_memory = False
debug_evaluator_mp = False
trim_debug_print = True

max_txt = 70*2


def set_debug_level(level):
    s = 1 if level == 0 else level/abs(level)
    globals()['debug_default_level'] = s*(abs(level) % 4)
    globals()['debug_essential_bc'] = abs(level) & (1 << 2) != 0
    globals()['debug_memory'] = abs(level) & (1 << 3) != 0


def dprint(*args, **kargs):
    s = ''
    for item in args:
        s = s + ' ' + str(item)

    if trim_debug_print and "notrim" not in kargs:
        s = textwrap.shorten(s, width=max_txt, placeholder='...')

    if debug_mode != 0:
        import sys
        print('DEBUG('+str(debug_mode)+')::'+s)


def find_by_id(_id_):
    '''
    find an object using id 
    '''
    import gc
    for obj in gc.get_objects():
        if id(obj) == _id_:
            return obj
    raise Exception("No found")


class DPrint(object):
    def __init__(self, name, level):
        self.name = name
        self.level = level

    def __call__(self, *args, **kargs):
        if 'stack' in kargs:
            traceback.print_stack()
        s = ''

        from petram.mfem_config import use_parallel
        if use_parallel:
            from mpi4py import MPI
            myid = MPI.COMM_WORLD.rank
        else:
            myid = 0

        for item in args:
            s = s + ' ' + str(item)

        if trim_debug_print and "notrim" not in kargs:
            s = textwrap.shorten(s, width=max_txt, placeholder='...')

        if self.name in debug_modes:
            if debug_modes[self.name] >= self.level:
                print('DEBUG('+str(self.name)+' ' + str(myid)+')::'+s)
        else:
            if debug_default_level < 0:
                if abs(debug_default_level) >= self.level:
                    print('DEBUG('+str(self.name)+' ' + str(myid)+')::'+s)
            else:
                if (abs(debug_default_level) >= self.level and
                        myid == 0):
                    print('DEBUG('+str(self.name)+' ' + str(myid)+')::'+s)


class RPrint(object):
    def __init__(self, name, head_only=False):
        self.name = name
        self.head_only = head_only

    def __call__(self, *args, **kargs):
        if 'stack' in kargs:
            traceback.print_stack()
        s = ''
        try:
            from mpi4py import MPI
            myid = MPI.COMM_WORLD.rank
        except ImportError:
            myid = 0

        if self.head_only and myid != 0:
            return
        for item in args:
            s = s + ' ' + str(item)

        if trim_debug_print and "notrim" not in kargs:
            s = textwrap.shorten(s, width=max_txt, placeholder='...')
        print(str(self.name)+'(' + str(myid)+')::'+s)


def regular_print(n, head_only=False):
    return RPrint(n, head_only)


def prints(n):
    return DPrint(n, 1), DPrint(n, 2), DPrint(n, 3)


def set_level(name, level):
    debug_modes[name] = level


def init_dprints(name, level=None):
    if level is not None:
        set_level(name, level)
    return prints(name)


def format_memory_usage(point="memory usage"):
    usage = resource.getrusage(resource.RUSAGE_SELF)
    return '''%s: usertime=%s systime=%s mem=%s mb
           ''' % (point, usage[0], usage[1], memory_usage_resource())
#                (usage[2]*resource.getpagesize())/1000000.0 )


def memory_usage_resource():
    import resource
    import sys
    rusage_denom = 1024.
    if sys.platform == 'darwin':
        # ... it seems that in OSX the output is different units ...
        rusage_denom = rusage_denom * rusage_denom
    mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / rusage_denom
    return mem


try:
    import guppy
    hasGUPPY = True
except ImportError:
    hasGUPPY = False

if hasGUPPY:
    def format_heap__usage():
        from guppy import hpy
        h = hpy()
        return h.heap()
else:
    def format_heap__usage():
        pass


def use_profiler(method):
    def method2(self, *args, **kwargs):
        if self.use_profiler:
            import cProfile
            import pstats
            from six import StringIO
            pr = cProfile.Profile()
            pr.enable()

        val = method(self, *args, **kwargs)

        if self.use_profiler:
            pr.disable()
            s = StringIO()
            sortby = 'cumulative'
            ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
            ps.print_stats()
            # print(s.getvalue())

            from petram.mfem_config import use_parallel
            if use_parallel:
                from mpi4py import MPI
                myid = MPI.COMM_WORLD.rank
                smyid = '.''{:0>6d}'.format(myid)
            else:
                smyid = ''
            ps.dump_stats("cProfile_"+self.name()+".out"+smyid)
        return val
    return method2


def flush_stdout(method):
    def method2(self, *args, **kwargs):
        import sys
        sys.stdout.flush()
        sys.stderr.flush()
        val = method(self, *args, **kwargs)
        return val

    return method2


class ConvergenceError(Exception):
    """Base class for exceptions in this module."""
    pass

    def __init__(self, message="Failed to converge"):
        self.message = message


'''
a decorator to time a method
'''


def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        print("entering:", method.__name__.upper())
        log = kw.pop('log_time', None)
        result = method(*args, **kw)
        te = time.time()
        if log is not None:
            name = log.get('log_name', method.__name__.upper())
            log[name] = int((te - ts) * 1000)
        else:
            print('%r  %2.2f ms' % (method.__name__, (te - ts) * 1000))
        print("done:", method.__name__.upper())
        return result
    return timed


def handle_allow_python_function_coefficient(message):
    from petram.mfem_config import get_allow_python_function_coefficient

    dprint1 = DPrint("Python mode check", 1)

    if get_allow_python_function_coefficient() == "warn":
        import traceback
        traceback.print_stack()
        dprint1(message)
    elif get_allow_python_function_coefficient() == "ignore":
        pass
    elif get_allow_python_function_coefficient() == "always use Python coeff.":
        pass
    elif get_allow_python_function_coefficient() == "error":
        assert False, message




import numpy as np


def do_connect_pairs(ll):
    d = {}

    flags = [False, ]*len(ll)
    count = 1
    d[ll[0][0]] = ll[0][1]
    flags[0] = True
    while count < len(ll):
        for k, f in enumerate(flags):
            if f:
                continue
            l = ll[k]
            if l[0] in d.keys():
                d[l[1]] = l[0]
                flags[k] = True
                count = count + 1
                break
            elif l[1] in d.keys():
                d[l[0]] = l[1]
                flags[k] = True
                count = count + 1
                break
            else:
                pass
        else:
            break
    key = list(d)[0]
    pt = [key]
    lmax = len(d)
    while d[key] != pt[0]:
        pt.append(d[key])
        key = d[key]
        if len(pt) > lmax:
            break
    if d[key] == pt[0]:
        pt.append(pt[0])

    ll = [l for l, f in zip(ll, flags) if not f]
    return pt, ll


def connect_pairs(ll):
    ret = []
    rest = ll
    while len(rest) != 0:
        pt, rest = do_connect_pairs(rest)
        ret.append(pt)
    if len(ret) == 1:
        return pt
    return ret


def do_find_circle_center(p1, p2, p3,  norm):
    p = p2 - p1
    q = np.cross(norm, p)
    d = np.sqrt(np.sum(p**2))

    p = p/np.sqrt(np.sum(p**2))
    q = q/np.sqrt(np.sum(q**2))

    a = np.sum((p3-p1)*p)
    b = np.sum((p3-p1)*q)

    m = np.linalg.inv(np.array([[0, b/2.], [1., -a/2.]]))
    v = np.array([[a/2.-d/2.], [b/2.]])

    s, t = np.dot(m, v)
    c = p1 + (p2-p1)/2.+s[0]*q
    return c


def find_circle_center_radius(vv, norm):
    '''
    assuming that point are somewhat equally space, I will
    scatter three points around the circle given by input points

    this approch reduces the risk of having three points colinear,
    Note that we need to have sufficently large number of points...

    '''
    k = len(vv)-2
    ii = np.linspace(0, len(vv)-1, 4).astype(int)

    pts = [do_find_circle_center(vv[i+ii[0]], vv[i+ii[1]], vv[i+ii[2]], norm)
           for i in range(ii[1]-ii[0])]

    ctr = np.mean(pts, 0)
    r = np.mean(np.sqrt(np.sum((vv - ctr)**2, 1)))
    return ctr, r


def connect_pairs2(ll):
    '''
    connect paris of indices to make a loop

    (example)
    >>> idx = array([[1, 4],  [3, 4], [1,2], [2, 3],])
    >>> connect_pairs(idx)
    [[1, 2, 3, 4, 1]]
    '''
    if not isinstance(ll, np.ndarray):
        ll = np.array(ll)

    idx = np.where(ll[:, 0] > ll[:, 1])[0]
    t1 = ll[idx, 0]
    t2 = ll[idx, 1]
    ll[idx, 0] = t2
    ll[idx, 1] = t1

    ii = np.vstack([np.arange(ll.shape[0]), ]*2).transpose()
    d = np.ones(ll.shape[0]*ll.shape[1]).reshape(ll.shape)
    from scipy.sparse import csr_matrix, coo_matrix
    m = coo_matrix((d.flatten(), (ii.flatten(), ll.flatten())),
                   shape=(len(ll), np.max(ll+1)))
    mm = m.tocsc()
    ic = mm.indices
    icp = mm.indptr
    mm = m.tocsr()
    ir = mm.indices
    irp = mm.indptr

    def get_start(taken):
        idx = np.where(np.logical_and(np.diff(icp) == 1, taken == 0))[0]
        nz = np.where(np.logical_and(np.diff(icp) != 0, taken == 0))[0]
        if len(nz) == 0:
            return
        if len(idx) > 0:
            #print('Open end found')
            pt = (ic[icp[idx[0]]], idx[0])
        else:
            pt = (ic[icp[nz[0]]], nz[0])
        pts = [pt]
        return pts

    def hop_v(pt):
        ii = pt[1]
        ii = [icp[ii], icp[ii+1]-1]
        next = ic[ii[1]] if ic[ii[0]] == pt[0] else ic[ii[0]]
        return (next, pt[1])

    def hop_h(pt):
        ii = pt[0]
        ii = [irp[ii], irp[ii+1]-1]
        next = ir[ii[1]] if ir[ii[0]] == pt[1] else ir[ii[0]]
        return (pt[0], next)

    def trace(pts):
        loop = [pts[-1][1]]
        while True:
            pts.append(hop_v(pts[-1]))
            # rows.append(pts[-1][0])
            pts.append(hop_h(pts[-1]))
            if pts[-1][1] in loop:
                break  # open ended
            loop.append(pts[-1][1])
            if pts[-1] == pts[0]:
                break
        return loop

    taken = (icp*0)[:-1]
    loops = []
    while True:
        pts = get_start(taken)
        if pts is None:
            break
        loop = trace(pts)
        loops.append(loop)
        taken[loop] = 1

    return loops


def find_crosssetional_shape(mesh, a, b, c, d):
    '''

    inspect mesh and find geometry represents crosssection of mesh
    slice by a plane 
       ax + by + cz + d = 0
    returns
       return shapes, shapes_attr, edge2point, coords

       shape : [(array of edges),,,]
       shape_attr : domain attribute
       edge2point : edges of shape
       coords : node point of edges.

       note: shape is mixture of triangle, quad, -- hex.

    '''
    from petram.mfem_config import use_parallel
    if use_parallel:
        import mfem.par as mfem
    else:
        import mfem.ser as mfem
    v = mfem.Vector()
    mesh.GetVertices(v)
    vv = v.GetDataArray().copy()
    vv = vv.reshape(3, -1)
    val = vv[0, :]*a + vv[1, :]*b + vv[2, :]*c + d

    attrs = mesh.GetAttributeArray()

    class number_generator(object):
        def __init__(self):
            self._d = {}
            self.num = -1

        def get_number(self, p):
            try:
                if p[0] < p[1]:
                    pp = (p[0], p[1])
                else:
                    pp = (p[1], p[0])
            except:
                pp = p

            if pp in self._d:
                return self._d[pp], False
            else:
                self.num = self.num+1
                self._d[pp] = self.num
                return self.num, True

        def get_reverse_map(self):
            return {self._d[p]: p for p in self._d}

    shapes = []
    shapes_attr = []
    shape_ptx = []

    Nv = number_generator()  # vertex numbers
    Ne = number_generator()  # edge numbers

    face_done = np.zeros(mesh.GetNFaces(), dtype=bool)
    face_cache = {}

    for iele in range(mesh.GetNE()):
        inodes = mesh.GetElementVertices(iele)
        ifaces, void = mesh.GetElementFaces(iele)

        f = val[inodes]

        if np.all(f >= 0) or np.all(f <= 0):
            if np.sum(f == 0) < 3:
                continue  # element is one side of plane.
                # only edge/node may be on the plane.
            # face is on the plane
            ifaces = [x for x in ifaces if not face_done[x]]

            for iface in ifaces:
                face_done[iface] = True
                f = val[mesh.GetFaceVertices(iface)]

                if np.all(f == 0):
                    iedges, orts = mesh.GetFaceEdges(iface)
                    shape = []
                    for iedge, ort in zip(iedges, orts):
                        iv = mesh.GetEdgeVertices(iedge)
                        n1, isNew = Nv.get_number(iv[0])
                        if isNew:
                            shape_ptx.append(vv[:, iv[0]])
                        n2, isNew = Nv.get_number(iv[1])
                        if isNew:
                            shape_ptx.append(vv[:, iv[1]])

                        e1, isNew = Ne.get_number((n1, n2))

                        shape.append(e1)
                    shapes.append(shape)
                    shapes_attr.append(attrs[iele])
                    break
                else:
                    pass

        else:
            # face intersect with the plane
            shape = []
            for iface in ifaces:
                if iface in face_cache:
                    e1 = face_cache[iface]
                    shape.append(e1)
                else:
                    iedges = mesh.GetFaceEdges(iface)[0]
                    edge0 = []
                    for iedge in iedges:
                        iv = mesh.GetEdgeVertices(iedge)
                        iv1, iv2 = iv
                        f = val[iv]
                        if np.max(f) * np.min(f) < 0:
                            ptx = (vv[:, iv1]*val[iv2] - vv[:, iv2]
                                   * val[iv1])/(- val[iv1] + val[iv2])
                            idx = (iv1, iv2)
                        elif f[0] == 0:
                            ptx = vv[:, iv1]
                            idx = iv1
                        elif f[1] == 0:
                            ptx = vv[:, iv2]
                            idx = iv2
                        else:
                            continue
                        n, isNew = Nv.get_number(idx)
                        if n in edge0:
                            continue
                        if isNew:
                            shape_ptx.append(ptx)
                        edge0.append(n)
                        c = c+1
                        if c == 1:
                            break
                    if len(edge0) < 2:
                        continue
                    e1, isNew = Ne.get_number(edge0)
                    shape.append(e1)
                    face_cache[iface] = e1
            shapes.append(shape)
            shapes_attr.append(attrs[iele])

    edge2point = Ne.get_reverse_map()
    coords = np.vstack(shape_ptx)
    return shapes, shapes_attr, edge2point, coords


def find_cp_pc_parameter(mesh, abcd, e1, gsize=None, gcount=100, origin=None, attrs=None):
    '''
    make cut-plane point_clund

    cp is defined by norm (a, b, c) and d as ax+by+cz+d = 0
    e1 is one axes on cp
    e2 is computed automatically

    gsize : grid size

    '''
    from petram.mfem_config import use_parallel
    if use_parallel:
        import mfem.par as mfem
    else:
        import mfem.ser as mfem

    v = mfem.Vector()
    mesh.GetVertices(v)
    sdim = mesh.SpaceDimension()
    dim = mesh.Dimension()
    vv = v.GetDataArray().copy()

    vv = vv.reshape(sdim, -1)
    if sdim == 3:
        pass
    elif sdim == 2:
        vv = np.vstack([vv, vv[:1, :]*0])
    else:
        assert False, "1D mesh is not supported"

    norm = np.array(abcd[:3])
    norm = norm/np.sqrt(np.sum(norm**2))

    if sdim == 3 and dim == 3:
        # flag elements which are on the plane
        flag = np.zeros(mesh.GetNV(), dtype=bool)
        for iele, f in enumerate(mesh.IsElementOnPlaneArray(*abcd)):
            if not f:
                continue
            flag[mesh.GetElementVertices(iele)] = True
    else:
        flag = np.ones(mesh.GetNV(), dtype=bool)
    '''
    val =  vv[0, :]*abcd[0] + vv[1,:]*abcd[1] + vv[2,:]*abcd[2] + abcd[3]
    for iele in range(mesh.GetNE()):
        inodes = mesh.GetElementVertices(iele)
        f = val[inodes]
        
        if np.all(f > 0) or np.all(f < 0):
            continue

        if attrs is not None and not (mesh.GetAttribute(iele) in attrs):
            continue
        
        flag[mesh.GetElementVertices(iele)] = True
    '''

    points = np.transpose(vv[:, flag])

    p1 = np.sum(points*abcd[:3], 1)+abcd[3]
    dd = np.sqrt(np.sum(np.array(abcd[:3])**2))
    p1 = p1/dd
    points = points - norm.reshape(3)*p1.reshape(-1, 1)

    e2 = np.cross(norm, e1)
    e1 = np.cross(e2, norm)
    e1 = e1 / np.sqrt(np.sum(e1**2))
    e2 = e2 / np.sqrt(np.sum(e2**2))

    if len(points) == 0:
        return None

    origin = points[0] if origin is None else np.array(origin)

    x = np.sum((points-origin)*e1, -1)
    y = np.sum((points-origin)*e2, -1)

    xmax = np.max(x)
    xmin = np.min(x)
    ymax = np.max(y)
    ymin = np.min(y)

    dx = xmax-xmin
    dy = ymax-ymin

    if gsize is None:
        try:
            gsize1 = dx/(gcount[0]-1)
            gsize2 = dx/(gcount[1]-1)
        except:
            gsize1 = max(dx/(gcount-1), dy/(gcount-1))
            gsize2 = gsize1
    else:
        try:
            gsize1 = gsize[0]
            gsize2 = gsize[1]
        except:
            gsize1 = gsize
            gsize2 = gsize

    return {"origin": origin, "e1": e1, "e2": e2,
            "x": (xmin, xmax, gsize1), "y": (ymin, ymax, gsize2)}


def generate_pc_from_cpparam(origin=None, e1=None, e2=None, x=None, y=None):
    xmin, xmax, xsize = x
    ymin, ymax, ysize = y
    xx = np.linspace(xmin, xmax, int((xmax-xmin)/xsize))
    yy = np.linspace(ymin, ymax, int((ymax-ymin)/ysize))

    XX, YY = np.meshgrid(xx, yy)
    e1 = np.atleast_2d(e1)
    e2 = np.atleast_2d(e2)
    pc = np.array(origin) + e1*XX.reshape(-1, 1) + e2*YY.reshape(-1, 1)
    pc = pc.reshape(len(yy), len(xx), 3)
    return pc


def make_cp_pc(mesh, abcd, e1, gsize=None, gcount=100):
    param = find_cp_pc_parameter(mesh, abcd, e1, gsize=gsize, gcount=gcount)
    pc = generate_pc_from_cpparam(param)
    return pc



from ifigure.interactive import figure

def open_meshviewer():
    viewer = figure(); viewer.threed(True)
    viewer.xlabel('x')
    viewer.ylabel('y')
    viewer.zlabel('z')
    w = viewer.attach_to_main()    
    return w



import numpy as np
import scipy

import petram.debug as debug
debug.debug_default_level = 1
dprint1, dprint2, dprint3 = debug.init_dprints('dof_map')

from petram.mfem_config import use_parallel
if use_parallel:
   from mpi4py import MPI
   comm = MPI.COMM_WORLD
   num_proc = MPI.COMM_WORLD.size
   myid     = MPI.COMM_WORLD.rank
   from petram.helper.mpi_recipes import *
   import mfem.par as mfem
   from mfem.common.mpi_debug import nicePrint
   
else:
   import mfem.ser as mfem
   num_proc = 1
   myid = 0
   def allgather(x): return [x]
   nicePrint = dprint1

'''
def map_dom2dom():


def map_dom2bdr():


def map_bdr2bdr():
'''

def find_element_ctr(mesh, idx, bdr):
    if len(idx) == 0:
        sdim = mesh.SpaceDimension()
        return np.empty(shape=(0,sdim))
    if bdr:
        m = mesh.GetBdrElementVertices
    else:
        m = mesh.GetElementVertices

    pts = np.mean(np.array([[mesh.GetVertexArray(i) for i in m(k)]
                             for k in idx]), 1)
    return pts
    

def find_element(mesh, use_bdr, attrs):
    if use_bdr:
        arr = mesh.GetBdrAttributeArray()
    else:
        arr = mesh.GetAttributeArray()
    idx = np.arange(len(arr))
        
    if attrs!='all':
        try:
            _void = attrs[0]
        except:
            attrs = [attrs]
        flag = np.in1d(arr, attrs)
        idx = idx[flag]
    return idx

def get_dombdr_flag(dim1, dim2, map_bdr):
    if dim1 == dim2:
        map_bdr1 = map_bdr
        map_bdr2 = map_bdr
    elif dim1 == dim2+1:
        map_bdr1 = map_bdr
        if not map_bdr:
            assert False, "can not map domain to a mesh with lower dim"
        else:
            map_bdr2 = False
    elif dim1 == dim2-1:
        map_bdr1 = map_bdr
        if not map_bdr:
            assert False, "can not map boundary to a mesh with higher dim"
        else:
            map_bdr2 = True
    else:
        assert False, "can not create element mapping between meshes"

    return map_bdr1, map_bdr2

def map_element(mesh1, mesh2, attr='all', map_bdr=False):
    '''
    create mapping between element
    '''

    dim1 = mesh1.Dimension()
    dim2 = mesh2.Dimension()

    map_bdr1, map_bdr2 = get_dombdr_flag(dim1, dim2, map_bdr)    
        
    el_arr1 = find_element(mesh1, map_bdr1, attr)
    el_arr2 = find_element(mesh2, map_bdr2, attr)

    ct1 = find_element_ctr(mesh1, el_arr1, map_bdr1)
    ct2 = find_element_ctr(mesh2, el_arr2, map_bdr2)    

                    
    if use_parallel:
        # share ibr2 (destination information among nodes...)
        ct1dim = ct1.shape[1] if ct1.size > 0 else 0
        ct1dim = comm.allgather(ct1dim)
        ct1 = np.atleast_2d(ct1).reshape(-1, max(ct1dim))
        ct2 = np.atleast_2d(ct2).reshape(-1, max(ct1dim))
        ct2_arr = [comm.bcast(ct2, root=i) for i in range(num_proc)]
    else:
        ct2_arr = [ct2]       
       
    from scipy.spatial import cKDTree

    dists = np.zeros(len(el_arr1))+np.inf

    nodes = np.zeros(len(el_arr1), dtype=int)
    el = np.zeros(len(el_arr1), dtype=int)-1
    num_target = 0
    for rank, ct2 in enumerate(ct2_arr):
        tree = cKDTree(ct2)
        ctr_dist, map_1_2 = tree.query(ct1)
        kk = (ctr_dist < dists)
        dists[kk] = ctr_dist[kk]
        nodes[kk] = rank
        el[kk] = map_1_2[kk]
        num_target += len(ct2)

    ## remove data whose distance is too big.
    ## need to find a threshold by gathering dist data.
    if use_parallel:
        dists_all = allgather_vector(dists)
    else:
        dists_all = dists

    sel = dists <= np.sort(dists_all)[num_target-1]
    el_arr1 = el_arr1[sel]
    el = el[sel]
    nodes = nodes[sel]
    
    el_map = {}
    for rank in range(num_proc):
        kk = (nodes == rank)
        el_map[rank] = dict(zip(el_arr1[kk], el[kk]))

    # reverse map
    rev_el_map = {}    
    if use_parallel:
        for i in range(num_proc):
            tmp = comm.gather(dict(zip(el_map[i].values(), list(el_map[i]))), root=i)
            if i == myid:
                data = tmp
                
        for rank in range(num_proc):
            rev_el_map[rank] = data[rank]
    else:
        rev_el_map[0] = dict(zip(el_map[0].values(), list(el_map[0])))
        
    '''
    format:
       el_map[mpi_rank] = {element_id in mesh1: element_id_mesh2}
         # elemnt_id_mesh1 is mapped in mpi_rank as elemet_id_mesh2
       el_mapr[mpi_rank] = {element_id_mesh2: element_id in mesh1}
         # elemnt_id_mesh2 is mapped in mpi_rank as elemet_id_mesh1
    '''
    return el_map, rev_el_map

    



'''

   Distance module

   physics module to compute distance

'''
import numpy as np

from petram.model import Domain, Bdry, Point, Pair
from petram.phys.phys_model import Phys, PhysModule

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('Distance_Model')

txt_predefined = ''
model_basename = 'Distance'


class Distance_DefDomain(Domain, Phys):
    can_delete = False

    def __init__(self, **kwargs):
        super(Distance_DefDomain, self).__init__(**kwargs)

    def get_panel1_value(self):
        return None

    def import_panel1_value(self, v):
        pass


class Distance_DefBdry(Bdry, Phys):
    can_delete = False
    is_essential = False

    def __init__(self, **kwargs):
        super(Distance_DefBdry, self).__init__(**kwargs)
        Phys.__init__(self)

    def attribute_set(self, v):
        super(Distance_DefBdry, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = ['remaining']
        return v

    def get_possible_bdry(self):
        return []


class Distance_DefPoint(Point, Phys):
    can_delete = False
    is_essential = False

    def __init__(self, **kwargs):
        super(Distance_DefPoint, self).__init__(**kwargs)
        Phys.__init__(self)

    def attribute_set(self, v):
        super(Distance_DefPoint, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = ['']
        return v


class Distance_DefPair(Pair, Phys):
    can_delete = False
    is_essential = False
    is_complex = False

    def __init__(self, **kwargs):
        super(Distance_DefPair, self).__init__(**kwargs)
        Phys.__init__(self)

    def attribute_set(self, v):
        super(Distance_DefPair, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v


class Distance(PhysModule):
    dim_fixed = False

    def __init__(self, **kwargs):
        super(Distance, self).__init__()
        Phys.__init__(self)
        #self['Domain'] = Distance_DefDomain()
        self['Boundary'] = Distance_DefBdry()
        self['Point'] = Distance_DefPoint()
        #self['Pair'] = Distance_DefPair()

    @property
    def dep_vars(self):
        ret = self.dep_vars_base
        ret = [x + self.dep_vars_suffix for x in ret]
        if self.generate_grad_fespace:
            ret = ret + ["vec" + x for x in ret]
        return ret

    @property
    def dep_vars_base(self):
        val = self.dep_vars_base_txt.split(',')
        return val

    @property
    def dep_vars0(self):
        val = self.dep_vars_base_txt.split(',')
        return [x + self.dep_vars_suffix for x in val]

    @property
    def der_vars(self):
        if self.generate_grad_fespace:
            sdim = self.geom_dim
            dd = self.dep_vars[0]
            names = ['vec' + dd + txt.strip()
                     for txt in self.ind_vars.split(',')]
        else:
            names = []
        return names

    @property
    def vdim(self):
        return [1, self.ndim]

    @vdim.setter
    def vdim(self, val):
        pass

    def get_fec_type(self, idx):
        t = ['H1', 'H1v'+str(self.ndim)]
        return t[idx]

    def postprocess_after_add(self, engine):
        try:
            sdim = engine.meshes[0].SpaceDimension()
        except:
            return
        if sdim == 3:
            self.ind_vars = 'x, y, z'
            self.ndim = 3
        elif sdim == 2:
            self.ind_vars = 'x, y'
            self.ndim = 2
        elif sdim == 1:
            self.ind_vars = 'x'
            self.ndim = 1
        else:
            pass

    def is_complex(self):
        return self.is_complex_valued

    def get_fec(self):
        v = self.dep_vars
        el = [self.element,
              'H1_FECollection']
        return [(vv, ee) for vv, ee in zip(v, el)]

    def attribute_set(self, v):
        v = super(Distance, self).attribute_set(v)
        v["element"] = 'H1_FECollection'
        v["dim"] = 2
        v["ind_vars"] = 'x, y'
        v["dep_vars_suffix"] = ''
        v["dep_vars_base_txt"] = 'dist'
        v["is_complex_valued"] = False
        v["generate_grad_fespace"] = False
        return v

    def panel1_param(self):
        sdim = self.geom_dim

        import wx
        panels = super(Distance, self).panel1_param()
        a, b = self.get_var_suffix_var_name_panel()
        panels.extend([
            ["independent vars.", self.ind_vars, 0, {}],
            a, b,
            ["derived vars.", ','.join(self.der_vars), 2, {}],
            ["predefined ns vars.", txt_predefined, 2, {}],
            ["generate vector", self.generate_grad_fespace, 3, {"text": ' '}], ])
        return panels

    def get_panel1_value(self):
        names = self.dep_vars_base_txt
        names2 = ', '.join(self.der_vars)
        val = super(Distance, self).get_panel1_value()

        val.extend([self.ind_vars,
                    self.dep_vars_suffix,
                    names, names2, txt_predefined,
                    self.generate_grad_fespace])
        return val

    def import_panel1_value(self, v):
        import ifigure.widgets.dialog as dialog

        v = super(Distance, self).import_panel1_value(v)
        self.ind_vars = str(v[0])
        self.is_complex_valued = False
        self.dep_vars_suffix = str(v[1])
        self.element = "H1_FECollection"
        self.dep_vars_base_txt = ','.join(
            [x.strip() for x in str(v[2]).split(',')])
        self.generate_grad_fespace = bool(v[5])
        return True

    '''
    The same as parent
    def get_possible_domain(self):
        from petram.phys.wf.wf_constraints import WF_WeakDomainBilinConstraint, WF_WeakDomainLinConstraint
        return [WF_WeakDomainBilinConstraint, WF_WeakDomainLinConstraint]
    '''

    def get_possible_bdry(self):
        from petram.phys.distance.wall import Distance_Wall

        bdrs = super(Distance, self).get_possible_bdry()
        return [Distance_Wall]

    def get_possible_point(self):
        '''
        To Do. Support point source
        '''
        return []

    def get_possible_pair(self):
        return []

    def add_variables(self, v, name, solr, soli=None):
        from petram.helper.variables import add_coordinates
        from petram.helper.variables import add_scalar
        from petram.helper.variables import add_components
        from petram.helper.variables import add_expression
        from petram.helper.variables import add_surf_normals
        from petram.helper.variables import add_constant
        from petram.helper.variables import GFScalarVariable

        ind_vars = [x.strip() for x in self.ind_vars.split(',')]
        suffix = self.dep_vars_suffix

        #from petram.helper.variables import TestVariable
        #v['debug_test'] =  TestVariable()

        add_coordinates(v, ind_vars)
        add_surf_normals(v, ind_vars)

        dep_vars = self.dep_vars
        sdim = self.geom_dim

        if name == dep_vars[0]:
            add_scalar(v, name, "", ind_vars, solr, soli)

        elif name == dep_vars[1]:
            for k, suffix in enumerate(ind_vars):
                nn = name + suffix
                v[nn] = GFScalarVariable(solr, soli, comp=k+1)
        else:
            pass
        return v



import numpy as np
import scipy
import six
import weakref
from weakref import WeakKeyDictionary as WKD
from weakref import WeakValueDictionary as WVD


from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

from petram.sol.evaluator_agent import EvaluatorAgent
from petram.sol.bdr_nodal_evaluator import process_iverts2nodals
from petram.sol.bdr_nodal_evaluator import eval_at_nodals, get_emesh_idx


class PointcloudEvaluator(EvaluatorAgent):
    def __init__(self, attrs, pc_type=None, pc_param=None):
        '''
           attrs = [1,2,3]
           plane = [ax, ay, az, c]

           cut-plane is defined as
           ax * x + ay * y + ax * z + c = 0
        '''
        super(PointcloudEvaluator, self).__init__()
        self.points = None
        self.subset = None
        self.attrs = set(attrs)
        self.pc_type = pc_type
        self.pc_param = pc_param

    def preprocess_geometry(self, attrs, emesh_idx=0, pc_type=None,
                            pc_param=None):

        from petram.helper.geom import generate_pc_from_cpparam

        self.attrs = attrs
        if pc_param is not None:
            pc_param = self.pc_param
            pc_type = self.pc_type

        if pc_type == 'cutplane':  # cutplane
            param = {"origin": pc_param[0], "e1": pc_param[1], "e2": pc_param[2],
                     "x": pc_param[3], "y": pc_param[4]}
            cp_abc = np.cross(pc_param[1], pc_param[2])
            cp_d = -np.sum(cp_abc*pc_param[0])
            points = generate_pc_from_cpparam(**param)

        elif pc_type == 'line':
            sp = np.array(pc_param[0])
            ep = np.array(pc_param[1])
            num = pc_param[2]

            ii = np.linspace(0, 1., num)
            points = np.vstack([sp * (1-i) + ep * i for i in ii])

        elif pc_type == 'xyz':
            points = pc_param

        mesh = self.mesh()[emesh_idx]
        sdim = mesh.SpaceDimension()

        if points.shape[-1] > sdim:
            points = points[..., :sdim]

        if np.prod(points.shape) == 0:
            assert False, "PointCloud: Number of points = 0"
            return

        self.ans_shape = points.shape
        self.ans_points = points
        self.points = points.reshape(-1, points.shape[-1])

        v = mfem.Vector()
        mesh.GetVertices(v)
        vv = v.GetDataArray()

        vv = vv.reshape(sdim, -1)
        max_mesh_ptx = np.max(vv, 1)
        min_mesh_ptx = np.min(vv, 1)

        max_ptx = np.max(self.points, 0)
        min_ptx = np.min(self.points, 0)

        out_of_range = False

        for i in range(len(max_mesh_ptx)):
            if max_mesh_ptx[i] < min_ptx[i]:
                out_of_range = True
            if min_mesh_ptx[i] > max_ptx[i]:
                out_of_range = True

        self.subset = None

        if pc_type == "cutplane" and sdim == 3:
            # in 3D, we try to cut down the number of point query to FindPoints
            param = vv[0, :]*cp_abc[0] + vv[1, :] * \
                cp_abc[1] + vv[2, :]*cp_abc[2] + cp_d
            if np.max(param)*np.min(param) == 0:
                out_of_range = True

            else:
                x1 = np.sum((vv.transpose() - pc_param[0])*pc_param[1], -1)
                xxx = np.min(x1), np.max(x1)
                y1 = np.sum((vv.transpose() - pc_param[0])*pc_param[2], -1)
                yyy = np.min(y1), np.max(y1)

                xmin, xmax, xsize = pc_param[3]
                ymin, ymax, ysize = pc_param[4]
                xxx = [int((xxx[0]-xmin)//xsize), int((xxx[1]-xmin)//xsize)]
                yyy = [int((yyy[0]-ymin)//ysize), int((yyy[1]-ymin)//ysize)]

                ss = self.ans_points.shape

                if xxx[0] < 0:
                    xxx[0] = 0
                if yyy[0] < 0:
                    yyy[0] = 0
                if xxx[1] >= ss[1]:
                    xxx[1] = ss[1]-1
                if yyy[1] >= ss[0]:
                    yyy[1] = ss[0]-1
                if xxx[0] > 0:
                    xxx[0] = xxx[0] - 1
                if yyy[0] > 0:
                    yyy[0] = yyy[0] - 1
                if xxx[1] < self.ans_points.shape[1]-1:
                    xxx[1] = xxx[1] + 1
                if yyy[1] < self.ans_points.shape[0]-1:
                    yyy[1] = yyy[1] + 1

                subset = np.zeros(ss[:-1], dtype=bool)
                subset[yyy[0]:yyy[1], xxx[0]:xxx[1]] = True

                ptx = self.ans_points[yyy[0]:yyy[1], :, :]
                ptx = ptx[:, xxx[0]:xxx[1], :]
                self.points = ptx.reshape(-1, self.ans_points.shape[-1])
                self.subset = subset

        if out_of_range:
            counts = 0
            elem_ids = np.zeros(len(self.points), dtype=int)-1
            int_points = [None]*len(self.points)
            print("skipping mesh")
        else:
            print("Chekcing " + str(len(self.points)) + " points")
            counts, elem_ids, int_points = mesh.FindPoints(
                self.points, warn=False)
            print("FindPoints found " + str(counts) + " points")
        attrs = [mesh.GetAttribute(id) if id != -1 else -1 for id in elem_ids]
        attrs = np.array([i if i in self.attrs else -1 for i in attrs])

        elem_ids = [-1 if a == -1 else eid for a, eid in zip(attrs, elem_ids)]
        counts = np.sum(np.array(elem_ids) != -1)

        self.elem_ids = elem_ids
        self.masked_attrs = attrs

        self.int_points = int_points
        self.counts = counts

        idx = np.where(attrs != -1)[0]
        self.locs = self.points[idx]

        self.valid_idx = idx
        self.emesh_idx = emesh_idx
        self.knowns = WKD()

    def eval_at_points(self, expr, solvars, phys):
        from petram.helper.variables import (Variable,
                                             var_g,
                                             NativeCoefficientGenBase,
                                             CoefficientVariable,
                                             NumbaCoefficientVariable)

        variables = []
        code = compile(expr, '<string>', 'eval')
        names = code.co_names

        g = {}
        # print solvars.keys()
        for key in phys._global_ns.keys():
            g[key] = phys._global_ns[key]
        for key in solvars.keys():
            g[key] = solvars[key]

        ll_name = []
        ll_value = []
        var_g2 = var_g.copy()

        new_names = []
        name_translation = {}

        all_names = list(names[:])

        def get_names(names):
            for n in names:
                if (n in g and isinstance(g[n], Variable)):
                    new_names = g[n].get_names()
                    for x in new_names:
                        all_names.append(x)
                    get_names(new_names)
        get_names(names)

        for n in all_names:
            if (n in g and isinstance(g[n], NativeCoefficientGenBase)):
                g[n+"_coeff"] = CoefficientVariable(g[n], g)
                new_names.append(n+"_coeff")
                name_translation[n+"_coeff"] = n

            elif (n in g and isinstance(g[n], NumbaCoefficientVariable)):
                ind_vars = [xx.strip() for xx in phys.ind_vars.split(',')]
                g[n].set_coeff(ind_vars, g)
                new_names.append(n)
                name_translation[n] = n

            elif (n in g and isinstance(g[n], Variable)):
                new_names.extend(g[n].dependency)
                new_names.append(n)
                name_translation[n] = n

            elif n in g:
                new_names.append(n)
                name_translation[n] = n

        for n in new_names:
            if (n in g and isinstance(g[n], Variable)):
                if not g[n] in self.knowns:
                    self.knowns[g[n]] = g[n].point_values(counts=self.counts,
                                                          locs=self.locs,
                                                          attrs=self.masked_attrs,
                                                          elem_ids=self.elem_ids,
                                                          mesh=self.mesh()[
                                                              self.emesh_idx],
                                                          int_points=self.int_points,
                                                          g=g,
                                                          knowns=self.knowns)

                #ll[n] = self.knowns[g[n]]
                ll_name.append(name_translation[n])
                ll_value.append(self.knowns[g[n]])
            elif (n in g):
                var_g2[n] = g[n]

        if len(ll_value) > 0:
            val = np.array([eval(code, var_g2, dict(zip(ll_name, v)))
                            for v in zip(*ll_value)])
        else:
            # if expr does not involve Varialbe, evaluate code once
            # and generate an array
            val = np.array([eval(code, var_g2)]*len(self.locs))

        return val

    def eval(self, expr, solvars, phys):
        from petram.sol.bdr_nodal_evaluator import get_emesh_idx

        emesh_idx = get_emesh_idx(self, expr, solvars, phys)
        if len(emesh_idx) > 1:
            assert False, "expression involves multiple mesh (emesh length != 1)"

        if len(emesh_idx) == 1:
            if self.emesh_idx != emesh_idx[0]:
                self.preprocess_geometry(self.attrs, emesh_idx=emesh_idx[0],
                                         pc_type=self.pc_type,
                                         pc_param=self.pc_param)

        if self.counts == 0:
            return None, None, None

        val = self.eval_at_points(expr, solvars, phys)

        if val is None:
            return None, None, None

        shape = self.ans_shape[:-1]

        if self.subset is None:
            attrs = self.masked_attrs.reshape(shape)
        else:
            attrs = np.zeros(self.ans_shape[:-1], dtype=int)-1
            attrs[self.subset] = self.masked_attrs.reshape(
                self.points.shape[:-1])

        return self.ans_points, val, attrs



import numpy as np
from collections import defaultdict

from mfem.ser import GlobGeometryRefiner as GR
from mfem.ser import DenseMatrix
default_refine = 5

def get_geom(idx, size, base, get_transformation, attrs, sdim, refine = None):
    # base 
    # SEGMENT     = 1
    # TRIANGLE    = 2
    # SQUARE      =p 3
    # TETRAHEDRON = 4
    # CUBE        = 5
    if refine is None: refine = default_refine

    RefG = GR.Refine(base, refine)
    T = get_transformation(idx[0])
    ir = RefG.RefPts
    npt = ir.GetNPoints()
    ref_idx = np.array(RefG.RefGeoms.ToList()).reshape(-1, size)
    refe_idx = np.array(RefG.RefEdges.ToList()).reshape(-1, 2)
    ptx_x = np.zeros((npt*len(idx), sdim), dtype=np.float32)

    m = DenseMatrix()

    ptx_data = [None]*len(idx)
    for k, i in enumerate(idx):
        T = get_transformation(i)
        T.Transform(ir, m)
        ptx_data[k] = m.GetDataArray().copy()
    ptx_x = np.hstack(ptx_data).transpose().astype(np.float32, copy=False)

    '''
    for k, i in enumerate(idx):
        T = get_transformation(i)
        ptx = np.vstack([T.Transform(ir.IntPoint(j)) for j in range(npt)])
        ptx_x[k*npt:(k+1)*npt, :] = ptx.astype(np.float32, copy = False)

    '''
    ivert_x = (np.tile(ref_idx, (len(idx), 1)) +
             np.repeat(np.arange(len(idx))*npt,
                       len(ref_idx)).reshape(-1, 1))
    iverte_x = (np.tile(refe_idx, (len(idx), 1)) +
             np.repeat(np.arange(len(idx))*npt,
                       len(refe_idx)).reshape(-1, 1))

    if isinstance(attrs, dict):  # for edge
        a = [attrs[i] for i in idx]
    else:
        a = attrs[idx]
    attrs_x = np.repeat(a, npt)

    ivert_x = ivert_x.flatten().reshape(-1,size)
    attrs = a


    return attrs, ptx_x, ivert_x, iverte_x, attrs_x



'''
   extract_refined_mesh_data3

   extract refined  mesh data from 2D (=ndim)  mesh
'''
import numpy as np
from collections import defaultdict

from petram.mesh.find_edges import find_edges
from petram.mesh.find_vertex import find_vertex
from mfem.ser import GlobGeometryRefiner as GR

def extract_refined_mesh_data2(mesh, refine = None):
    
    ndim = mesh.Dimension()
    sdim = mesh.SpaceDimension()
    
    ivert0 = [mesh.GetElement(i).GetVerticesArray()
                           for i in range(mesh.GetNE())]
    attrs = mesh.GetAttributeArray()
    nvert = np.array([len(x) for x in ivert0])
    idx3 = np.where(nvert == 3)[0]
    idx4 = np.where(nvert == 4)[0]
    ivert = []; ivert3 = None; ivert4 = None
    iv3 = []; iv4 = []

    from refined_mfem_geom import get_geom

    ptx = []; ivx3 = None; 
    if len(idx3) != 0:
        base = mesh.GetElementBaseGeometry(idx3[0])
        gt = mesh.GetElementTransformation        
        attr3, ptx3, ivx3, ivxe3, attrx3 =  get_geom(idx3, 3, base,
                                                     gt, attrs, sdim, refine)
        ptx.append(ptx3)
        iv3 = [ivert0[k] for k in idx3]      
        ivert3 = np.vstack(iv3)
    if len(idx4) != 0:
        base = mesh.GetElementBaseGeometry(idx4[0])
        gt = mesh.GetElementTransformation                
        attr4, ptx4, ivx4, ivxe4, attrx4 =  get_geom(idx4, 4, base,
                                                          gt, attrs,sdim,refine)
        ptx.append(ptx4)
        iv4 = [ivert0[k] for k in idx4]
        ivert4 = np.vstack(iv4)
        if ivx3 is not None:
            ivx4 = ivx4 + len(ptx3)       
            ivxe4 = ivxe4 + len(ptx3)
        
    ptx_face = np.vstack(ptx)
    
    # unique edges for node                       
    tmp = np.hstack(iv3 + iv4)
    u, indices = np.unique(tmp, return_inverse=True)
    
    ll3 = 3*len(idx3)
    indices3 = indices[:ll3]
    indices4 = indices[ll3:]
                       
    table = np.zeros(np.max(u)+1, dtype=int)-1
    for k, u0 in enumerate(u):
        table[u0] = k

    X = np.vstack([mesh.GetVertexArray(k) for k in u])    
    if X.shape[1] == 2:
        X = np.hstack((X, np.zeros((X.shape[0],1))))
    elif X.shape[1] == 1:
        X = np.hstack((X, np.zeros((X.shape[0],2))))
    
    kdom = mesh.attributes.ToList()
    
    cells = {}
    cell_data = {}
    cell_data['X_refined_face']=ptx_face
    if ivert3 is not None:
        cells['triangle'] = indices3.reshape(ivert3.shape)
        cells['triangle_x'] = ivx3
        cells['triangle_xe'] = ivxe3        
        cell_data['triangle'] = {}
        cell_data['triangle']['physical'] = attr3
        cell_data['triangle_x'] = {}                         
        cell_data['triangle_x']['physical'] = attrx3
    if ivert4 is not None:
        cells['quad'] = indices4.reshape(ivert4.shape)
        cells['quad_x'] = ivx4
        cells['quad_xe'] = ivxe4        
        cell_data['quad'] = {}
        cell_data['quad']['physical'] = attr4
        cell_data['quad_x'] = {}                         
        cell_data['quad_x']['physical'] = attrx4                         
    
    ## fill line/surface loop info
    loop= {}
    for k in kdom:
        loop[k] = []
        
    for ibdr in range(mesh.GetNBE()):
        idx = mesh.GetBdrElementEdgeIndex(ibdr)
        elem1, elem2 = mesh.GetFaceElements(idx)
        if elem1 != -1:
            i = mesh.GetAttribute(elem1)
            loop[i].append(attrs[ibdr])
        if elem2 != -1:
            i = mesh.GetAttribute(elem2)
            loop[i].append(attrs[ibdr])
    for k in kdom:
        loop[k] = np.unique(loop[k])
        
    l_s_loop = loop, None
    
    line2edge = {}
    vert2vert = {}
    if mesh.GetNBE() == 0:
        # 2D surface mesh in 3D space could have no NBE
        iedge2bb = {}
        mesh.extended_connectivity = {'line2vert':None,
                                      'surf2line':l_s_loop[0],
                                      'vol2surf': l_s_loop[1],
                                      'line2edge':line2edge,
                                      'vert2vert':vert2vert}
        return X, cells, cell_data, l_s_loop, iedge2bb
    ## fill line
    battrs = mesh.GetBdrAttributeArray()
    ivert = np.vstack([mesh.GetBdrElement(i).GetVerticesArray()
                           for i in range(mesh.GetNBE())])
    cells['line'] = table[ivert]
    cell_data['line'] = {}
    cell_data['line']['physical'] = battrs
    iedge2bb = {k:k for k in battrs}

    base = 1
    gt = mesh.GetBdrElementTransformation
    idx2 = range(mesh.GetNBE())
    attr2, ptx2, ivx2, ivxe2, attrx2 = get_geom(idx2, 2, base, gt, battrs,
                                                sdim,refine)
    
    cells['line_x'] = ivx2
    cell_data['line_x'] = {}                        
    cell_data['line_x']['physical'] = attrx2
    cell_data['X_refined_edge']=ptx2
    
    ## points
    d = defaultdict(list)
    for i in range(mesh.GetNBE()):
        d[attrs[i]].extend(ivert[i, :])
    corners = {}
    for key in d:
       seen = defaultdict(int)
       for iiv in d[key]:
           seen[iiv] += 1
       corners[key] = [kk for kk in seen if seen[kk]==1]
    iverts = np.unique(np.hstack([corners[key] for key in corners]))
    
    cell_data['vertex'] = {}        
    if len(iverts) != 0:        
        cells['vertex'] = table[iverts]
        cell_data['vertex']['physical'] = np.arange(len(iverts))+1
    
    line2vert = {key: np.array(corners[key])+1 for key in corners}
    mesh.extended_connectivity = {'line2vert':line2vert,
                                  'surf2line':l_s_loop[0],
                                  'vol2surf': l_s_loop[1],
                                  'line2edge':line2edge,
                                  'vert2vert':vert2vert}
    
    ## iedge2bb : mapping from edge_id to boundary numbrer set
    ## X, cells, cell_data : the same data strucutre as pygmsh
    return X, cells, cell_data, l_s_loop, iedge2bb
